# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for brax.physics."""

import copy
import itertools
import os

from absl.testing import absltest
from absl.testing import parameterized
import brax
from brax import jumpy as jp
import jax

from google.protobuf import text_format


class BodyTest(absltest.TestCase):

  def test_projectile_motion(self):
    """A ball with an initial velocity curves down due to gravity."""
    sys = brax.System(
        text_format.Parse(
            """
    dt: 1 substeps: 1000
    gravity { z: -9.8 }
    bodies { name: "Ball" mass: 1 }
    defaults { qps { name: "Ball" vel {x: 1}}}
    dynamics_mode: "legacy_spring"
    """, brax.Config()))
    qp = sys.default_qp()
    qp, _ = sys.step(qp, jp.array([]))
    # v = v_0 + a * t
    self.assertAlmostEqual(qp.vel[0, 2], -9.8, 2)
    # x = x_0 + v_0 * t + 0.5 * a * t^2
    self.assertAlmostEqual(qp.pos[0, 0], 1, 2)
    self.assertAlmostEqual(qp.pos[0, 2], -9.8 / 2, 2)


class BoxTest(absltest.TestCase):

  _CONFIG = """
    dt: 1.5 substeps: 1000 friction: 0.77459666924 baumgarte_erp: 0.1
    gravity { z: -9.8 }
    bodies {
      name: "box" mass: 1
      colliders { box { halfsize { x: 0.5 y: 0.5 z: 0.5 }}}
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies { name: "Ground" frozen: { all: true } colliders { plane {}}}
    defaults { qps { name: "box" pos { z: 1 }}}
    defaults { qps { name: "box" pos { z: 2 } vel {x: 2}}}
    dynamics_mode: "legacy_spring"
  """

  def test_box_hits_ground(self):
    """A box falls onto the ground and stjp."""
    sys = brax.System(text_format.Parse(BoxTest._CONFIG, brax.Config()))
    qp = sys.default_qp(0)
    qp, _ = sys.step(qp, jp.array([]))
    self.assertAlmostEqual(qp.pos[0, 2], 0.5, 2)

  def test_box_slide(self):
    """A box slides across the ground and comes to a stop."""
    sys = brax.System(text_format.Parse(BoxTest._CONFIG, brax.Config()))
    qp = sys.default_qp(1)
    qp, _ = sys.step(qp, jp.array([]))
    self.assertAlmostEqual(qp.pos[0, 2], 0.5, 2)
    self.assertGreater(qp.pos[0, 0], 1)  # after sliding for a bit...
    self.assertAlmostEqual(qp.vel[0, 0], 0, 2)  # friction brings it to a stop
    self.assertLess(qp.pos[0, 0], 1.5)  # ... and keeps it from travelling 2m


class BoxBoxTest(absltest.TestCase):
  _CONFIG = """
    dt: 0.5 substeps: 200 friction: 0.8 elasticity: 0.5
    gravity { z: -9.8 }
    bodies {
      name: "box1" mass: 1
      colliders { box { halfsize { x: 0.2 y: 0.2 z: 0.2 }}}
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies {
      name: "box2" mass: 1
      colliders { box { halfsize { x: 0.1 y: 0.1 z: 0.1 }}}
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies { name: "Ground" frozen: { all: true } colliders { plane {}}}
    defaults {
      qps { name: "box1" pos { x: 0 y: 1 z: .2 } rot {z: 0} }
      qps { name: "box2" pos { x: 0.1 y: 1 z: .6 } rot {z: 45} }
    }
    dynamics_mode: "legacy_spring"
  """

  def test_box_box(self):
    """A box falls onto another box."""
    sys = brax.System(text_format.Parse(BoxBoxTest._CONFIG, brax.Config()))
    qp = sys.default_qp()
    qp, _ = sys.step(qp, jp.array([]))
    # Boxes are stacked.
    self.assertAlmostEqual(qp.pos[0, 2], 0.2, delta=0.03)
    self.assertAlmostEqual(qp.pos[1, 2], 0.5, delta=0.03)
    # x-y position of the top box is unchanged.
    self.assertAlmostEqual(qp.pos[1, 0], 0.1, delta=0.03)
    self.assertAlmostEqual(qp.pos[1, 1], 1.0, delta=0.03)


class CollisionDebuggerTest(absltest.TestCase):
  _CONFIG = """
    dt: 0.01 substeps: 4 friction: 1
    gravity { z: -9.8 }
    bodies {
      name: "box" mass: 1
      colliders { box { halfsize { x: 0.5 y: 0.5 z: 0.5 }}}
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies { name: "Ground" frozen: { all: true } colliders { plane {}}}
    defaults { qps { name: "box" pos { z: 0.49 }}}
    dynamics_mode: "legacy_spring"
  """

  def test_system_runs_with_debug_on(self):
    """Tests that the simulation runs with debug_contacts."""
    sys = brax.System(
        text_format.Parse(CollisionDebuggerTest._CONFIG, brax.Config()))
    qp = sys.default_qp(0)
    qp, info = sys.step(qp, jp.array([]))
    self.assertGreater(info.contact_pos.shape[0], 0)


class BoxCapsuleTest(absltest.TestCase):

  _CONFIG = """
    dt: 0.05 substeps: 20 friction: 1 baumgarte_erp: 0.1
    gravity { z: -9.8 }
    bodies {
      name: "box" mass: 1
      colliders { box { halfsize { x: 0.5 y: 0.5 z: 0.5 }}}
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies {
      name: "capsule" mass: 1
      colliders { capsule { length: 2 radius: 0.2 } }
      inertia { x: 1 y: 1 z: 1 }
    }

    bodies { name: "Ground" frozen: { all: true } colliders { plane {}}}
    defaults {
      qps { name: "box" pos { z: 2 }}
      qps { name: "capsule" pos: { z: 0.2 } rot: { y: 90 } }
    }
    dynamics_mode: "legacy_spring"
  """

  def test_box_hits_capsule(self):
    """A box falls onto a capsule and stays above it."""
    sys = brax.System(text_format.Parse(BoxCapsuleTest._CONFIG, brax.Config()))
    qp = sys.default_qp()
    self.assertAlmostEqual(qp.pos[0, 2], 2, 2)

    step = jax.jit(sys.step)
    for _ in range(50):
      qp, _ = step(qp, jp.array([]))
    # Box should be on the capsule, rather than on the ground.
    self.assertAlmostEqual(qp.pos[0, 2], 0.9, 2)


class HeightMapTest(absltest.TestCase):

  _CONFIG = """
    dt: 2 substeps: 1000 friction: 1 baumgarte_erp: 0.1 elasticity: 0
    gravity { z: -9.8 }
    bodies {
      name: "box" mass: 1
      colliders { box { halfsize { x: 0.3 y: 0.3 z: 0.3 }}}
      inertia { x: 0.1 y: 0.1 z: 0.1 }
    }
    bodies {
      name: "ground"
      frozen: { all: true }
      colliders {
        heightMap {
          size: 10
          data: [0, 0, 0, 0, 0, 0, 0, 0, 0]
        }
      }
    }
    defaults { qps { name: "box" pos: {x: 5 y: 5 z: 1}}}
    dynamics_mode: "legacy_spring"
  """

  def test_box_stays_on_heightmap(self):
    """A box falls onto the height map and stjp."""
    sys = brax.System(text_format.Parse(HeightMapTest._CONFIG, brax.Config()))
    qp = sys.default_qp()
    qp, _ = jax.jit(sys.step)(qp, jp.array([]))
    self.assertAlmostEqual(qp.pos[0, 2], 0.3, 2)


class SphereTest(absltest.TestCase):

  _CONFIG = """
    dt: 5 substeps: 50 friction: 0.6 baumgarte_erp: 0.1
    gravity { z: -9.8 }
    bodies {
      name: "Sphere1" mass: 1
      colliders { sphere { radius: 0.25}}
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies { name: "Ground" frozen: { all: true } colliders { plane {}}}
    defaults {qps { name: "Sphere1" pos {z: 1}}}
    defaults {qps { name: "Sphere1" pos {z: 1} vel {x: 2}}}
    dynamics_mode: "legacy_spring"
  """

  def test_sphere_hits_ground(self):
    """A sphere falls onto the ground and stops."""
    sys = brax.System(text_format.Parse(SphereTest._CONFIG, brax.Config()))
    qp = sys.default_qp(0)
    qp, _ = sys.step(qp, jp.array([]))
    self.assertAlmostEqual(qp.pos[0, 2], 0.25, 2)

  def test_sphere_roll(self):
    """A sphere rolls across the ground."""
    sys = brax.System(text_format.Parse(SphereTest._CONFIG, brax.Config()))
    qp = sys.default_qp(1)
    qp, _ = sys.step(qp, jp.array([]))
    self.assertGreater(qp.ang[0, 1], 0.25)  # sphere is rolling


class CapsuleTest(absltest.TestCase):

  _CONFIG = """
    dt: 20.0 substeps: 10000 friction: 0.6 baumgarte_erp: 0.1
    gravity { z: -9.8 }
    bodies {
      name: "Capsule1" mass: 1
      colliders { capsule { radius: 0.25 length: 1.0 }}
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies {
      name: "Capsule2" mass: 1
      colliders { rotation { y: 90 } capsule { radius: 0.25 length: 1.0 }}
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies {
      name: "Capsule3" mass: 1
      colliders { rotation { y: 45 } capsule { radius: 0.25 length: 1.0 }}
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies {
      name: "Capsule4" mass: 1
      colliders { rotation { x: 45 } capsule { radius: 0.25 length: 1.0 }}
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies { name: "Ground" frozen: { all: true } colliders { plane {}}}
    defaults {
      qps { name: "Capsule1" pos {z: 1}}
      qps { name: "Capsule2" pos {x:1 z: 1}}
      qps { name: "Capsule3" pos {x:3 z: 1}}
      qps { name: "Capsule4" pos {x:5 z: 1}}
    }
    defaults {
      qps { name: "Capsule1" pos {z: 1}}
      qps { name: "Capsule2" pos {z: 2}}
      qps { name: "Capsule3" pos {x:3 z: 1}}
      qps { name: "Capsule4" pos {x:5 z: 1}}
    }
    dynamics_mode: "legacy_spring"
  """

  def test_capsule_hits_ground(self):
    """A capsule falls onto the ground and stops."""
    sys = brax.System(text_format.Parse(CapsuleTest._CONFIG, brax.Config()))
    qp = sys.default_qp(0)
    qp, _ = jax.jit(sys.step)(qp, jp.array([]))
    self.assertAlmostEqual(qp.pos[0, 2], 0.5, 2)  # standing up and down
    self.assertAlmostEqual(qp.pos[1, 2], 0.25, 2)  # lying on its side
    self.assertAlmostEqual(qp.pos[2, 2], 0.25, 2)  # rolls to side from y rot
    self.assertAlmostEqual(qp.pos[3, 2], 0.25, 2)  # rolls to side from x rot

  def test_capsule_hits_capsule(self):
    """A capsule falls onto another capsule and balances on it."""
    config = text_format.Parse(CapsuleTest._CONFIG, brax.Config())
    config.dt = 2.0
    config.substeps = 1000
    sys = brax.System(config)
    qp = sys.default_qp(1)
    qp, _ = jax.jit(sys.step)(qp, jp.array([]))
    self.assertAlmostEqual(qp.pos[0, 2], 0.5, 2)  # standing up and down
    self.assertAlmostEqual(qp.pos[1, 2], 1.25, 2)  # lying on Capsule1

  def test_cull(self):
    """A capsule falls onto another capsule, with NN culling."""
    config = text_format.Parse(CapsuleTest._CONFIG, brax.Config())
    config.collider_cutoff = 1
    config.dt = 2.0
    config.substeps = 1000
    sys = brax.System(config)
    qp = sys.default_qp(1)
    qp, _ = jax.jit(sys.step)(qp, jp.array([]))
    self.assertAlmostEqual(qp.pos[0, 2], 0.5, 2)  # standing up and down
    self.assertAlmostEqual(qp.pos[1, 2], 1.25, 2)  # lying on Capsule1


class MeshTest(absltest.TestCase):

  _CONFIG = """
    dt: 0.05 substeps: 10 friction: 1.0 baumgarte_erp: 0.1
    gravity { z: -9.8 }
    bodies {
      name: "Mesh" mass: 1
      colliders { mesh { name: "Cylinder" scale: 0.1 } }
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies {
      name: "Capsule" mass: 1
      colliders { capsule { length: 2 radius: 0.2 } }
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies { name: "Ground" frozen: { all: true } colliders { plane {} } }
    defaults {
      # Initial position is high up in the air.
      qps { name: "Mesh" pos: {x: 0 y: 0 z: 1.5} }

      # Capsule is away from the cylinder.
      qps { name: "Capsule" pos: {x: 10 y: 0 z: 0.2} rot: {x: 0 y: 90 z: 0} }
    }
    mesh_geometries {
      name: "Cylinder"
      path: "cylinder.stl"
    }
    dynamics_mode: "legacy_spring"
  """

  def test_mesh_hits_ground(self):
    """A mesh falls onto the ground."""
    config = text_format.Parse(MeshTest._CONFIG, brax.Config())
    resource_path = os.path.join(absltest.get_default_test_srcdir(),
                                 'brax/tests/testdata')
    sys = brax.System(config, [resource_path])
    qp = sys.default_qp()
    # Cylinder should be up in the air.
    self.assertAlmostEqual(qp.pos[0, 2], 1.5, 2)

    step = jax.jit(sys.step)
    for _ in range(30):
      qp, _ = step(qp, jp.array([]))
    # Cylinder should be on the ground.
    self.assertAlmostEqual(qp.pos[0, 2], 0, 2)

  def test_mesh_hits_capsule(self):
    config = text_format.Parse(MeshTest._CONFIG, brax.Config())
    # Move the capsule under the cylinder.
    config.defaults[0].qps[1].pos.x = 0
    resource_path = os.path.join(absltest.get_default_test_srcdir(),
                                 'brax/tests/testdata')
    sys = brax.System(config, [resource_path])
    qp = sys.default_qp()
    self.assertAlmostEqual(qp.pos[0, 2], 1.5, 2)

    step = jax.jit(sys.step)
    for _ in range(30):
      qp, _ = step(qp, jp.array([]))
    # Cylinder should be on the capsule, rather than on the ground.
    self.assertAlmostEqual(qp.pos[0, 2], 0.38, 2)


class CapsuleClippedPlaneTest(absltest.TestCase):
  """Tests the capsule-clippedPlane collision function."""
  _CONFIG = """
    dt: 2 substeps: 1000 friction: 0.6
    gravity { z: -9.8 }
    bodies {
      name: "Sphere1" mass: 1
      colliders { sphere { radius: 0.5 } }
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies {
      name: "Sphere2" mass: 1
      colliders { sphere { radius: 0.5 } }
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies {
      name: "Sphere3" mass: 1
      colliders { sphere { radius: 0.5 } }
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies {
      name: "ClippedPlane" mass: 1
      colliders {
        clipped_plane { halfsize_x: 3 halfsize_y: 1 }
        position { z: 2 }
      }
      frozen { all: true }
    }
    bodies { name: "Ground" frozen: { all: true } colliders { plane {}}}
    defaults {
      qps { name: "Sphere1" pos { z: 3 } }
      qps { name: "Sphere2" pos { z: 3 x: -4 } }
      qps { name: "Sphere3" pos { z: 3 y: -2 } }
      qps { name: "ClippedPlane" pos { x: 0 } }
    }
    dynamics_mode: "legacy_spring"
  """

  def test_collision(self):
    """Tests collisions between spheres and a clipped plane."""
    config = text_format.Parse(CapsuleClippedPlaneTest._CONFIG, brax.Config())
    sys = brax.System(config)

    qp = sys.default_qp()
    step_fn = jax.jit(sys.step)
    qp, _ = step_fn(qp, jp.array([]))
    # sphere1 is on the clipped plane and sphere2/sphere3 are on the ground
    # plane
    self.assertAlmostEqual(qp.pos[0, 2], 2.5, 2)
    self.assertAlmostEqual(qp.pos[1, 2], 0.5, 2)
    self.assertAlmostEqual(qp.pos[2, 2], 0.5, 2)


class JointTest(parameterized.TestCase):

  _CONFIG = """
    substeps: 100000
    dt: .01
    gravity { z: -9.8 }
    bodies {
      name: "Anchor" frozen: { all: true } mass: 1
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies { name: "Bob" mass: 1 inertia { x: 1 y: 1 z: 1 }}
    joints {
      name: "Joint" parent: "Anchor" child: "Bob" stiffness: 10000
      child_offset { z: 1 }
      angle_limit { min: -180 max: 180 }
    }
    dynamics_mode: "legacy_spring"
  """

  @parameterized.parameters((2.0, 0.125, 0.0625), (5.0, 0.125, 0.03125),
                            (1.0, 0.0625, 0.1))
  def test_pendulum_period(self, mass, radius, vel):
    """A small spherical mass swings for approximately one period."""
    config = text_format.Parse(JointTest._CONFIG, brax.Config())

    # this length of time comes from the following:
    # inertia_about_anchor = mass * dist_to_anchor^2 + inertia_cm
    # dist_to_anchor = 1.0
    # inertia_cm = 2/5 * mass * radius^2 (solid sphere)
    # T = 2 * pi * sqrt(inertia_about_anchor / (2 * mass * g * dist_to_anchor))
    config.dt = 2 * jp.pi * jp.sqrt((.4 * radius**2 + 1.) / 9.8)
    config.bodies[1].mass = mass
    config.bodies[1].inertia.x = .4 * mass * radius**2
    config.bodies[1].inertia.y = .4 * mass * radius**2
    config.bodies[1].inertia.z = .4 * mass * radius**2
    sys = brax.System(config)

    # initializing system to have a small initial velocity and ang velocity
    # so that small angle approximation is valid
    qp = brax.QP(
        pos=jp.array([[0., 0., 0.], [0., 0., -1.]]),
        rot=jp.array([[1., 0., 0., 0.], [1., 0., 0., 0.]]),
        vel=jp.array([[0., 0., 0.], [0., vel, 0.]]),
        ang=jp.array([[0., 0., 0.], [vel, 0., 0.]]))
    qp, _ = jax.jit(sys.step)(qp, jp.array([]))
    self.assertAlmostEqual(qp.pos[1, 1], 0., 3)  # returned to the origin

  offsets = [-15, 15, -45, 45, -75, 75]
  axes = [
      [1, 0, 0],
      [0, 1, 0],
      [0, 0, 1],
      [1, 1, 0],
      [0, 1, 1],
      [1, 0, 1],
      [1, 1, 1],
  ]
  limits = [0, 1]

  @parameterized.parameters(itertools.product(offsets, axes, limits))
  def test_reference_offset(self, offset, axis, limit):
    """Construct joint and check that default_qp generates offsets correctly."""
    config = text_format.Parse(JointTest._CONFIG, brax.Config())

    # loop over different types of joints
    for l in range(3):
      # construct appropriate number of angle_limits for this joint
      if l == 0:
        a_l = config.joints[0].angle_limit[0]
      else:
        a_l = config.joints[0].angle_limit.add()
      # set angle default to either 0 or the offset value
      a_l.min = offset * limit
      a_l.max = offset * limit

      sys_default = brax.System(config)

      this_offset = offset * jp.array(axis)

      # duplicate config deeply
      rotated_config = copy.deepcopy(config)

      rotated_config.joints[0].reference_rotation.x = this_offset[0]
      rotated_config.joints[0].reference_rotation.y = this_offset[1]
      rotated_config.joints[0].reference_rotation.z = this_offset[2]

      # construct a new config with this reference offset applied
      sys_offset = brax.System(rotated_config)
      offset_qp = sys_offset.default_qp()

      # construct joint functions for default and offset systems
      qp_p = jp.take(offset_qp, 0)
      qp_c = jp.take(offset_qp, 1)
      joint_offset = jp.take(sys_offset.joints[0], 0)
      joint_default = jp.take(sys_default.joints[0], 0)

      # calculate joint angles as seen by the default or offset system
      _, angle_offset = joint_offset.axis_angle(qp_p, qp_c)
      _, angle_default = joint_default.axis_angle(qp_p, qp_c)
      angle_offset = (jp.array(angle_offset) / jp.pi) * 180
      angle_default = (jp.array(angle_default) / jp.pi) * 180
      num_offsets = angle_offset.shape[0]

      for a_o, a_d, t_o in zip(angle_offset, angle_default,
                               this_offset[:num_offsets]):
        if limit == 0:
          # default system sees part rotated by offset degrees
          self.assertAlmostEqual(a_d, t_o, 2)
          # offset system sees part at local 0.0
          self.assertAlmostEqual(a_o, 0.0, 2)
        if limit == 1:
          # when default angle is nonzero, there's not a clean relationship
          # between the default and offset frames anymore, because the offset
          # frame now has two rotations applied---1) the rotation that offsets
          # it relative to the parent, and then 2) the rotation placing it at
          # its default angle limit.  it should still be the case, though, that
          # the offset angle agrees with the default angle limit.
          self.assertAlmostEqual(a_o, offset, 2)


class Actuator1DTest(parameterized.TestCase):

  _CONFIG = """
    substeps: 80
    dt: 4.0
    gravity { z: -9.8 }
    bodies {
      name: "Anchor" frozen: { all: true } mass: 1
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies { name: "Bob" mass: 1 inertia { x: 1 y: 1 z: 1 }}
    joints {
      name: "Joint" parent: "Anchor" child: "Bob" stiffness: 5000
      child_offset { z: 1 }
      angle_limit { min: -180 max: 180 }
      angular_damping: 20.0
    }
    actuators {
      name: "Joint"
      joint: "Joint"
      strength: 150.0
      angle {}
    }
    defaults { qps { name: "Anchor" pos {z: 2}} qps { name: "Bob" pos {z: 1}}}
    dynamics_mode: "legacy_spring"
  """

  @parameterized.parameters(15., 30., 45., 90.)
  def test_1d_angle_actuator(self, target_angle):
    """A simple part actuates to a target angle."""
    config = text_format.Parse(Actuator1DTest._CONFIG, brax.Config())
    sys = brax.System(config=config)
    qp = sys.default_qp()
    qp, _ = sys.step(qp, jp.array([target_angle]))
    qp_p = jp.take(qp, 0)
    qp_c = jp.take(qp, 1)
    joint = jp.take(sys.joints[0], 0)
    _, (angle,) = joint.axis_angle(qp_p, qp_c)

    self.assertAlmostEqual(target_angle * jp.pi / 180, angle, 2)


class Actuator2DTest(parameterized.TestCase):

  _CONFIG = """
    substeps: 2000
    dt: 2.0
    gravity { z: -9.8 }
    bodies {
      name: "Anchor" frozen: { all: true } mass: 1
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies {
      name: "Bob" mass: 1 inertia { x: 1 y: 1 z: 1 }
    }
    joints {
      name: "Joint" parent: "Anchor" child: "Bob" stiffness: 10000
      child_offset { z: 1 }
      angle_limit { min: -180 max: 180 }
      angle_limit { min: -180 max: 180 }
      angular_damping: 200.0
    }
    actuators {
      name: "Joint"
      joint: "Joint"
      strength: 2000.0
      angle {}
    }
    defaults { qps { name: "Anchor" pos {z: 2}} qps { name: "Bob" pos {z: 1}}}
    dynamics_mode: "legacy_spring"
  """

  @parameterized.parameters((15., 30.), (-45., 80), (120, -60.), (-35., -52.))
  def test_2d_angle_actuator(self, target_angle_1, target_angle_2):
    """A simple part actuates 2d-angle actuator to two target angles."""
    config = text_format.Parse(Actuator2DTest._CONFIG, brax.Config())
    sys = brax.System(config=config)
    qp = sys.default_qp()
    qp, _ = jax.jit(sys.step)(qp, jp.array([target_angle_1, target_angle_2]))
    qp_p = jp.take(qp, 0)
    qp_c = jp.take(qp, 1)
    joint = jp.take(sys.joints[0], 0)
    _, angles = joint.axis_angle(qp_p, qp_c)
    self.assertAlmostEqual(target_angle_1 * jp.pi / 180, angles[0], 2)
    self.assertAlmostEqual(target_angle_2 * jp.pi / 180, angles[1], 2)


class Actuator3DTest(parameterized.TestCase):

  _CONFIG = """
    substeps: 8000
    dt: 20
    bodies {
      name: "Anchor" frozen: { all: true } mass: 1
      inertia { x: 1 y: 1 z: 1 }
    }
    bodies {
      name: "Bob" mass: 1 inertia { x: 1 y: 1 z: 1 }
      colliders { capsule { radius: 0.5 length: 2.0 } }
    }
    joints {
      name: "Joint" parent: "Anchor" child: "Bob" stiffness: 10000
      child_offset { z: 1 }
      angle_limit {
        min: -100
        max: 100
      }
      angle_limit {
        min: -100
        max: 100
      }
      angle_limit {
        min: -100
        max: 100
      }
      angular_damping: 180.0
      limit_strength: 2000.0
    }
    actuators {
      name: "Joint"
      joint: "Joint"
      strength: 40.0
      torque {}
    }
    defaults { qps { name: "Anchor" pos {z: 2}} qps { name: "Bob" pos {z: 1}}}
    dynamics_mode: "legacy_spring"
  """
  torques = [(1, 0, 0), (0, 1, 0), (0, 0, 1), (1, 1, 1)]

  @parameterized.parameters(((15, 15, 15), torques), ((35, 40, 75), torques),
                            ((80, 45, 30), torques))
  def test_3d_torque_actuator(self, limits, torques):
    """A simple part actuates 3d-torque actuator to its limits."""
    config = text_format.Parse(Actuator3DTest._CONFIG, brax.Config())
    for t in torques:
      for angle_limit, limit in zip(config.joints[0].angle_limit, limits):
        angle_limit.min = -limit
        angle_limit.max = limit

      sys = brax.System(config=config)
      qp = sys.default_qp()
      qp, _ = jax.jit(sys.step)(qp, jp.array(t))
      qp_p = jp.take(qp, 0)
      qp_c = jp.take(qp, 1)
      joint = jp.take(sys.joints[0], 0)
      _, angles = joint.axis_angle(qp_p, qp_c)
      angles = [a * 180 / jp.pi for a in angles]
      for angle, limit, torque in zip(angles, limits, t):
        if torque != 0:
          self.assertAlmostEqual(angle, limit, 1)  # actuated to target angle


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ARS networks."""

from typing import Tuple

from brax.training import networks
from brax.training import types
from brax.training.types import PRNGKey
import jax.numpy as jnp

ARSNetwork = networks.FeedForwardNetwork


def make_policy_network(
    observation_size: int,
    action_size: int,
    preprocess_observations_fn: types.PreprocessObservationFn = types
    .identity_observation_preprocessor,
) -> ARSNetwork:
  """Creates a policy network."""

  def apply(processor_params, policy_params, obs):
    obs = preprocess_observations_fn(obs, processor_params)
    return jnp.matmul(obs, policy_params)

  return ARSNetwork(
      init=lambda _: jnp.zeros((observation_size, action_size)), apply=apply)


def make_inference_fn(policy_network: ARSNetwork):
  """Creates params and inference function for the ARS agent."""

  def make_policy(params: types.PolicyParams) -> types.Policy:

    def policy(observations: types.Observation,
               unused_key_sample: PRNGKey) -> Tuple[types.Action, types.Extra]:
      return policy_network.apply(*params, observations), {}

    return policy

  return make_policy



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Force applied to bodies."""

from typing import List, Tuple, Union

from brax import jumpy as jp
from brax import pytree
from brax.physics import bodies
from brax.physics import config_pb2
from brax.physics.base import P, QP


@pytree.register
class Thruster:
  """Applies a force to a body."""

  def __init__(self, forces: List[config_pb2.Force], body: bodies.Body,
               act_index: List[Tuple[int, int]]):
    """Thruster applies linear force to a body given a 3d act array.

    Args:
      forces: list of forces (all of the same type) to batch together
      body: (batched) bodfies for this force to act upon
      act_index: indices from the act array that drive this force
    """
    self.body = jp.take(body, [body.index[f.body] for f in forces])
    self.strength = jp.array([f.strength for f in forces])
    self.act_index = jp.array(act_index)

  def apply_reduced(self, force: jp.ndarray) -> jp.ndarray:
    dvel = force * self.strength / self.body.mass
    return dvel, jp.zeros_like(dvel)

  def apply(self, qp: QP, force_data: jp.ndarray) -> P:
    """Applies a force to a body.

    Args:
      qp: State data for system
      force_data: Data specifying the force to apply to a body.

    Returns:
      dP: The impulses that result from apply a force to the body.
    """

    force_data = jp.take(force_data, self.act_index)
    dvel, dang = jp.vmap(type(self).apply_reduced)(self, force_data)

    # sum together all impulse contributions to all parts effected by forces
    dvel = jp.segment_sum(dvel, self.body.idx, qp.pos.shape[0])
    dang = jp.segment_sum(dang, self.body.idx, qp.pos.shape[0])

    return P(vel=dvel, ang=dang)


@pytree.register
class Twister:
  """Applies a torque to a body."""

  def __init__(self, forces: List[config_pb2.Force], body: bodies.Body,
               act_index: List[Tuple[int, int]]):
    """Twister applies torque to a single body.

    Args:
      forces: list of forces (all of the same type) to batch together
      body: (batched) bodfies for this force to act upon
      act_index: indices from the act array that drive this force
    """
    self.body = jp.take(body, [body.index[f.body] for f in forces])
    self.strength = jp.array([f.strength for f in forces])
    self.act_index = jp.array(act_index)

  def apply_reduced(self, torque: jp.ndarray) -> jp.ndarray:
    dang = torque * self.strength / self.body.mass
    return jp.zeros_like(dang), dang

  def apply(self, qp: QP, force_data: jp.ndarray) -> P:
    """Applies a force to a body.

    Args:
      qp: State data for system
      force_data: Data specifying the force to apply to a body.

    Returns:
      dP: The impulses that result from apply a force to the body.
    """
    force_data = jp.take(force_data, self.act_index)
    dvel, dang = jp.vmap(type(self).apply_reduced)(self, force_data)

    # sum together all impulse contributions to all parts effected by forces
    dvel = jp.segment_sum(dvel, self.body.idx, qp.pos.shape[0])
    dang = jp.segment_sum(dang, self.body.idx, qp.pos.shape[0])

    return P(vel=dvel, ang=dang)


def get(config: config_pb2.Config,
        body: bodies.Body) -> List[Union[Thruster, Twister]]:
  """Creates all forces given a config and actuators."""
  # by convention, force indices are after actuator indices
  # get the next available index after actuator indices
  dofs = {j.name: len(j.angle_limit) for j in config.joints}
  current_index = sum([dofs[a.joint] for a in config.actuators])

  thrusters, thruster_indices = [], []
  twisters, twister_indices = [], []
  for f in config.forces:
    act_index = tuple(range(current_index, current_index + 3))
    current_index += 3
    if f.WhichOneof('type') == 'thruster':
      thrusters.append(f)
      thruster_indices.append(act_index)
    elif f.WhichOneof('type') == 'twister':
      twisters.append(f)
      twister_indices.append(act_index)
    else:
      raise ValueError(f'unknown force type: {f.WhichOneof("type")}')

  forces = []
  if thrusters:
    forces.append(Thruster(thrusters, body, thruster_indices))
  if twisters:
    forces.append(Twister(twisters, body, twister_indices))

  return forces



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for brax.physics.geometry."""

from absl.testing import absltest
from absl.testing import parameterized
import brax
from brax import jumpy as jp
from brax.physics import geometry
import jax
import numpy as np
import scipy


def _get_rand_triangle_vertices(seed=None):
  if seed is not None:
    np.random.seed(seed)
  verts = np.random.randn(3, 3)
  return verts[0, :], verts[1, :], verts[2, :]


def _get_rand_line_segment(seed=None):
  if seed is not None:
    np.random.seed(seed)
  verts = np.random.randn(2, 3)
  return verts[0, :], verts[1, :]


def _get_rand_point(seed=None):
  if seed is not None:
    np.random.seed(seed)
  verts = np.random.randn(1, 3)
  return verts[0, :]


def _get_rand_convex_polygon(seed=None):
  if seed is not None:
    np.random.seed(seed)
  points = np.random.random((15, 2))
  hull = scipy.spatial.ConvexHull(points)
  points = points[hull.vertices]
  points = np.array([[p[0], p[1], 0] for p in points])
  return points


def _get_rand_dir(seed=None, r=None):
  if seed is not None:
    np.random.seed(seed)
  r = r or np.random.randn(1)
  theta = np.random.random(1) * 2 * np.pi
  a = (np.random.random(1) - 0.5) * 2.0
  phi = np.arccos(a)
  x = r * np.sin(phi) * np.cos(theta)
  y = r * np.sin(phi) * np.sin(theta)
  z = r * np.cos(phi)
  return jp.array([x, y, z]).squeeze()


def _scipy_closest_triangle_point(p0, p1, p2, pt):
  # Parametrize the triangle using barycentric coordinates where a point in
  # the triangle Q = w * V_0 + y * V_1 + v * V_2
  # and u + v + w = 1, or w = 1 - (u + v), (u + v) <= 1.
  def fn(t):
    u, v = t[0], t[1]
    best_q = u * p0 + v * p1 + (1 - u - v) * p2
    return (pt - best_q).dot(pt - best_q)

  constraint = scipy.optimize.NonlinearConstraint(
      lambda x: [x[0], x[1], 1 - x[0] - x[1]],
      lb=np.array([0, 0, 0]),
      ub=np.array([1, 1, 1]))
  u, v = scipy.optimize.minimize(
      fn, jp.zeros(2), tol=1e-10, constraints=[constraint]).x
  best_q = u * p0 + v * p1 + (1 - u - v) * p2
  return best_q


def _scipy_closest_segment_to_segment_points(a0, a1, b0, b1):
  dir_a = a1 - a0
  len_a = jp.sqrt(dir_a.dot(dir_a))
  half_len_a = len_a / 2
  dir_a = dir_a / len_a

  dir_b = b1 - b0
  len_b = jp.sqrt(dir_b.dot(dir_b))
  half_len_b = len_b / 2
  dir_b = dir_b / len_b

  a_mid = a0 + dir_a * half_len_a
  b_mid = b0 + dir_b * half_len_b

  # Parametrize both line segments.
  def fn(t):
    best_a = a_mid + dir_a * t[0]
    best_b = b_mid + dir_b * t[1]
    return (best_a - best_b).dot(best_a - best_b)

  constraint = scipy.optimize.NonlinearConstraint(
      lambda x: x,
      lb=np.array([-half_len_a, -half_len_b]),
      ub=np.array([half_len_a, half_len_b]))
  ta, tb = scipy.optimize.minimize(
      fn, np.zeros(2), tol=1e-10, constraints=[constraint]).x
  best_a = a_mid + dir_a * ta
  best_b = b_mid + dir_b * tb
  return best_a, best_b


def _scipy_closest_segment_triangle_points(a, b, p0, p1, p2):
  # Parametrize line segment.
  dir_ls = b - a
  len_ls = jp.sqrt(dir_ls.dot(dir_ls))
  half_len_ls = len_ls / 2
  dir_ls = dir_ls / len_ls
  ls_mid = a + dir_ls * half_len_ls

  # Parametrize the triangle using barycentric coordinates.
  def fn(t):
    best_ls = ls_mid + dir_ls * t[0]
    u, v = t[1], t[2]
    best_q = u * p0 + v * p1 + (1 - u - v) * p2
    return (best_ls - best_q).dot(best_ls - best_q)

  constraint = scipy.optimize.NonlinearConstraint(
      lambda x: [x[0], x[1], x[2], 1 - x[1] - x[2]],
      lb=np.array([-half_len_ls, 0, 0, 0]),
      ub=np.array([half_len_ls, 1, 1, 1]))
  t_ls, u, v = scipy.optimize.minimize(
      fn, jp.zeros(3), tol=1e-10, constraints=[constraint]).x
  best_ls = ls_mid + dir_ls * t_ls
  best_q = u * p0 + v * p1 + (1 - u - v) * p2
  return best_ls, best_q


def _check_eq_pts(pts1, pts2, atol=1e-6):
  # For every point in pts1, make sure we have a point in pts2
  # that is close within `atol`, and vice versa.
  if not pts1.size and not pts2.size:
    return True
  elif not pts1.size:
    return False
  elif not pts2.size:
    return False
  if pts1.shape[-1] != 3 or pts2.shape[-1] != 3:
    raise AssertionError('Points should be three dimensional.')
  eq = True
  for p1 in pts1:
    eq = eq and jp.any(jp.sum(jp.abs(p1 - pts2) < atol, axis=-1) == 3)
  for p2 in pts2:
    eq = eq and jp.any(jp.sum(jp.abs(p2 - pts1) < atol, axis=-1) == 3)
  return eq


def _clip(clipping_polygon, subject_polygon):
  """Returns the clipped subject polygon, using Sutherland-Hodgman Clipping."""
  polygon1 = clipping_polygon
  polygon1_normal = jp.cross(polygon1[-1] - polygon1[0],
                             polygon1[1] - polygon1[0])
  clipping_edges = [
      (polygon1[i - 1], polygon1[i]) for i in range(len(polygon1))
  ]
  # Clipping plane normals point away from the clipping poly center (polygon
  # points assumed to have a clockwise winding order).
  clipping_planes = [
      (e0, jp.cross(polygon1_normal, e1 - e0)) for e0, e1 in clipping_edges
  ]

  output_polygon = subject_polygon
  for clipping_plane in clipping_planes:
    input_ = output_polygon
    output_polygon = []

    starting_pt = input_[-1]
    clipping_plane_pt, clipping_plane_normal = clipping_plane

    for endpt in input_:
      intersection_pt = brax.physics.geometry.closest_segment_point_plane(
          starting_pt, endpt, clipping_plane_pt, clipping_plane_normal)
      starting_pt_front = geometry.point_in_front_of_plane(
          clipping_plane_pt, clipping_plane_normal, starting_pt)
      endpt_front = geometry.point_in_front_of_plane(clipping_plane_pt,
                                                     clipping_plane_normal,
                                                     endpt)

      if not endpt_front and not starting_pt_front:
        output_polygon.append(endpt)
      elif not endpt_front and starting_pt_front:
        output_polygon.append(intersection_pt)
        output_polygon.append(endpt)
      elif not starting_pt_front:
        output_polygon.append(intersection_pt)

      starting_pt = endpt

    if not output_polygon:
      # All clipping points outside the subject polygon.
      return jp.array(output_polygon)

  return jp.array(output_polygon)


class ClosestSegmentPointsTest(absltest.TestCase):
  """Tests for closest segment points."""

  def test_closest_segments_points(self):
    a0 = jp.array([0.73432405, 0.12372768, 0.20272314])
    a1 = jp.array([1.10600128, 0.88555209, 0.65209485])
    b0 = jp.array([0.85599262, 0.61736299, 0.9843583])
    b1 = jp.array([1.84270939, 0.92891793, 1.36343326])
    best_a, best_b = geometry.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [1.09063, 0.85404, 0.63351], 5)
    self.assertSequenceAlmostEqual(best_b, [0.99596, 0.66156, 1.03813], 5)

  def test_intersecting_segments(self):
    """Tests segments that intersect."""
    a0, a1 = jp.array([0., 0., -1.]), jp.array([0., 0., 1.])
    b0, b1 = jp.array([-1., 0., 0.]), jp.array([1., 0., 0.])
    best_a, best_b = geometry.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [0., 0., 0.], 5)
    self.assertSequenceAlmostEqual(best_b, [0., 0., 0.], 5)

  def test_intersecting_lines(self):
    """Tests that intersecting lines (not segments) get clipped."""
    a0, a1 = jp.array([0.2, 0.2, 0.]), jp.array([1., 1., 0.])
    b0, b1 = jp.array([0.2, 0.4, 0.]), jp.array([1., 2., 0.])
    best_a, best_b = geometry.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [0.3, 0.3, 0.], 2)
    self.assertSequenceAlmostEqual(best_b, [0.2, 0.4, 0.], 2)

  def test_parallel_segments(self):
    """Tests that parallel segments have closest points at the midpoint."""
    a0, a1 = jp.array([0., 0., -1.]), jp.array([0., 0., 1.])
    b0, b1 = jp.array([1., 0., -1.]), jp.array([1., 0., 1.])
    best_a, best_b = geometry.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [0., 0., 0.], 5)
    self.assertSequenceAlmostEqual(best_b, [1., 0., 0.], 5)

  def test_parallel_offset_segments(self):
    """Tests that offset parallel segments are close at segment endpoints."""
    a0, a1 = jp.array([0., 0., -1.]), jp.array([0., 0., 1.])
    b0, b1 = jp.array([1., 0., 1.]), jp.array([1., 0., 3.])
    best_a, best_b = geometry.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [0., 0., 1.], 5)
    self.assertSequenceAlmostEqual(best_b, [1., 0., 1.], 5)

  def test_zero_length_segments(self):
    """Test that zero length segments don't return NaNs."""
    a0, a1 = jp.array([0., 0., -1.]), jp.array([0., 0., -1.])
    b0, b1 = jp.array([1., 0., 0.1]), jp.array([1., 0., 0.1])
    best_a, best_b = geometry.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [0., 0., -1.], 5)
    self.assertSequenceAlmostEqual(best_b, [1., 0., 0.1], 5)

  def test_overlapping_segments(self):
    """Tests that perfectly overlapping segments intersect at the midpoints."""
    a0, a1 = jp.array([0., 0., -1.]), jp.array([0., 0., 1.])
    b0, b1 = jp.array([0., 0., -1.]), jp.array([0., 0., 1.])
    best_a, best_b = geometry.closest_segment_to_segment_points(a0, a1, b0, b1)
    self.assertSequenceAlmostEqual(best_a, [0., 0., 0.], 5)
    self.assertSequenceAlmostEqual(best_b, [0., 0., 0.], 5)


class GeometryScipyTest(absltest.TestCase):
  """Tests for geometry functions against scipy equivalents."""

  def test_closest_triangle_point(self):
    for i in range(10):
      for j in range(10):
        pt = _get_rand_point(i)
        p0, p1, p2 = _get_rand_triangle_vertices(j)
        expected = _scipy_closest_triangle_point(p0, p1, p2, pt)
        ans = geometry.closest_triangle_point(p0, p1, p2, pt)
        self.assertSequenceAlmostEqual(expected, ans, 4)

  def test_closest_segment_to_segment_points(self):
    for i in range(10):
      for j in range(10):
        a0, a1 = _get_rand_line_segment(i)
        b0, b1 = _get_rand_line_segment(j)
        expected = _scipy_closest_segment_to_segment_points(a0, a1, b0, b1)
        ans = geometry.closest_segment_to_segment_points(a0, a1, b0, b1)
        expected_dist = (expected[0] - expected[1]).dot(expected[0] -
                                                        expected[1])
        test_dist = (ans[0] - ans[1]).dot(ans[0] - ans[1])
        self.assertAlmostEqual(expected_dist, test_dist, 4)

  def test_closest_segment_triangle_points(self):
    for i in range(10):
      for j in range(10):
        a, b = _get_rand_line_segment(i)
        p0, p1, p2 = _get_rand_triangle_vertices(j)
        expected = _scipy_closest_segment_triangle_points(a, b, p0, p1, p2)
        triangle_normal = brax.math.normalize(jp.cross(p0 - p1, p2 - p1))
        ans = geometry.closest_segment_triangle_points(a, b, p0, p1, p2,
                                                       triangle_normal)
        expected_dist = (expected[0] - expected[1]).dot(expected[0] -
                                                        expected[1])
        test_dist = (ans[0] - ans[1]).dot(ans[0] - ans[1])
        self.assertAlmostEqual(expected_dist, test_dist, 4)


class ClippingTest(parameterized.TestCase):
  """Tests that the clipping algorithm matches a baseline implementation."""
  clip_vectorized = jax.jit(geometry.clip)

  @parameterized.parameters(range(100))
  def test_clipped_triangles(self, i):
    subject_poly = jp.array(_get_rand_triangle_vertices(i))
    clipping_poly = jp.array(_get_rand_triangle_vertices(i + 1))

    poly_out = _clip(clipping_poly, subject_poly)

    clipping_normal = jp.cross(clipping_poly[-1] - clipping_poly[0],
                               clipping_poly[1] - clipping_poly[0])
    subject_normal = jp.cross(subject_poly[-1] - subject_poly[0],
                              subject_poly[1] - subject_poly[0])
    poly_out_jax, mask = ClippingTest.clip_vectorized(clipping_poly,
                                                      subject_poly,
                                                      clipping_normal,
                                                      subject_normal)

    self.assertTrue(
        _check_eq_pts(poly_out, poly_out_jax[mask], atol=1e-4),
        f'Clipped triangles {i} did not match.')

  @parameterized.parameters(range(100))
  def test_clipped_hulls(self, i):
    # The hulls are all in the x-y plane, unlike in `test_clipped_triangles`.
    subject_poly = jp.array(_get_rand_convex_polygon(i))
    clipping_poly = jp.array(_get_rand_convex_polygon(i + 1))

    poly_out = _clip(clipping_poly, subject_poly)

    clipping_normal = jp.cross(clipping_poly[-1] - clipping_poly[0],
                               clipping_poly[1] - clipping_poly[0])
    subject_normal = jp.cross(subject_poly[-1] - subject_poly[0],
                              subject_poly[1] - subject_poly[0])
    poly_out_jax, mask = ClippingTest.clip_vectorized(clipping_poly,
                                                      subject_poly,
                                                      clipping_normal,
                                                      subject_normal)

    self.assertTrue(
        _check_eq_pts(poly_out, poly_out_jax[mask], atol=1e-2),
        f'Clipped hulls {i} did not match.')


class OrthoganlsTest(parameterized.TestCase):
  """Tests the `get_orthogonals` function."""

  @parameterized.parameters(range(100))
  def test_orthogonals(self, i):
    a = _get_rand_dir(i)
    b, c = geometry.get_orthogonals(a)
    self.assertAlmostEqual(jp.abs(a.dot(b)), 0)
    self.assertAlmostEqual(jp.abs(b.dot(c)), 0)
    self.assertAlmostEqual(jp.abs(a.dot(c)), 0)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-import-not-at-top
"""Tools for exporting brax policies to other frameworks."""

from collections import abc
from typing import Any, Callable
import warnings


def _fix_frozen(d):
  """Changes any mappings (e.g. frozendict) back to dict."""
  if isinstance(d, list):
    return [_fix_frozen(v) for v in d]
  elif isinstance(d, tuple):
    return tuple(_fix_frozen(v) for v in d)
  elif not isinstance(d, abc.Mapping):
    return d
  d = dict(d)
  for k, v in d.items():
    d[k] = _fix_frozen(v)
  return d


def to_tf_model(path: str, inference_fn: Callable[..., Any], *trace_args):
  """Exports a brax inference function to a tensorflow saved model."""
  try:
    from jax.experimental import jax2tf
    import tensorflow as tf
  except ImportError:
    warnings.warn("""to_tf_model requires tensorflow.  Please run
    `pip install tensorflow` for this function to work.""")
    raise
  model = tf.Module()
  model.f = tf.function(jax2tf.convert(inference_fn), autograph=False)
  # saved_model doesn't recognize flax FrozenDict - convert them back to dict
  trace_args = _fix_frozen(trace_args)
  # for input tracing so that the model has the correct shapes
  model.f(*trace_args)
  tf.saved_model.save(model, path)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Exports a system config and trajectory as an html view."""

import os
from typing import List, Optional

import brax
from brax.io.file import File
from brax.io.file import MakeDirs
from brax.io.json import dumps


def save_html(path: str,
              sys: brax.System,
              qps: List[brax.QP],
              make_dir: bool = False):
  """Saves trajectory as a HTML file."""
  if make_dir and path:
    MakeDirs(os.path.dirname(path))
  with File(path, 'w') as fout:
    fout.write(render(sys, qps))


def render(sys: brax.System,
           qps: List[brax.QP],
           height: int = 480,
           info: Optional[brax.Info] = None) -> str:
  """Returns an HTML page that visualizes the system and qps trajectory."""
  if any((len(qp.pos.shape), len(qp.rot.shape)) != (2, 2) for qp in qps):
    raise RuntimeError('unexpected shape in qp.')
  system = dumps(sys, qps, info)
  html = _HTML.replace('<!-- system json goes here -->', system)
  html = html.replace('<!-- viewer height goes here -->', f'{height}px')
  return html


_HTML = """
<html>
  <head>
    <title>brax visualizer</title>
    <style>
      body {
        margin: 0;
        padding: 0;
      }
      #brax-viewer {
        margin: 0;
        padding: 0;
        height: <!-- viewer height goes here -->;
      }
    </style>
  </head>
  <body>
    <script type="application/javascript">
    var system = <!-- system json goes here -->;
    </script>
    <div id="brax-viewer"></div>
    <script type="module">
      import {Viewer} from 'https://cdn.jsdelivr.net/gh/google/brax@v0.1.0/js/viewer.js';
      const domElement = document.getElementById('brax-viewer');
      var viewer = new Viewer(domElement, system);
    </script>
  </body>
</html>
"""



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Joints connect bodies and constrain their movement.

These joint definitions are from the original Brax release, and constrain joints
via acceleration-level, springy updates.  We find these types of joints to be
simple to implement, fairly fast to simulate, but difficult to tune, typically
requiring significant iteration on the damping constants to get acceptable
motion quality. All future environments will be built with the newer,
position-level updates defined in joints.py which are infinitely stiff by
default.
"""

from typing import Any, List, Tuple

from brax import jumpy as jp
from brax import math
from brax import pytree
from brax.physics import bodies
from brax.physics import config_pb2
from brax.physics import joints
from brax.physics.base import P, QP, vec_to_arr


class Joint(joints.Joint):
  """A spring joint connects two bodies and constrains their movement.

  This constraint is determined by axes that define how to bodies may move in
  relation to one-another.
  """

  __pytree_ignore__ = ('index', 'dof')

  def __init__(self,
               joints: List[config_pb2.Joint],
               body: bodies.Body,
               spring_damping_coeff: float = 2.0):
    """Creates a Joint that connects two bodies and constrains their movement.

    Args:
      joints: list of joints (all of the same type) to batch together
      body: batched body that contain the parents and children of each joint
      spring_damping_coeff: coefficient for setting default spring damping
    """
    # springy-joint specific fields
    self.stiffness = jp.array([j.stiffness for j in joints])
    self.spring_damping = jp.array([
        j.spring_damping if j.HasField('spring_damping') else
        spring_damping_coeff * jp.sqrt(j.stiffness) for j in joints
    ])
    self.limit_strength = jp.array([
        j.limit_strength if j.HasField('limit_strength') else j.stiffness
        for j in joints
    ])
    self.angular_damping = jp.array([j.angular_damping for j in joints])
    self.limit = jp.array([[[i.min, i.max]
                            for i in j.angle_limit]
                           for j in joints]) / 180.0 * jp.pi
    self.body_p = jp.take(body, [body.index[j.parent] for j in joints])
    self.body_c = jp.take(body, [body.index[j.child] for j in joints])
    self.off_p = jp.array([vec_to_arr(j.parent_offset) for j in joints])
    self.off_c = jp.array([vec_to_arr(j.child_offset) for j in joints])
    self.index = {j.name: i for i, j in enumerate(joints)}
    self.dof = len(joints[0].angle_limit)
    v_rot = jp.vmap(math.rotate, include=[True, False])
    relative_quats = jp.array(
        [math.euler_to_quat(vec_to_arr(j.reference_rotation)) for j in joints])
    self.axis_c = jp.array([
        v_rot(jp.eye(3), math.euler_to_quat(vec_to_arr(j.rotation)))
        for j in joints
    ])
    self.axis_p = jp.array(
        [v_rot(j, r) for j, r in zip(self.axis_c, relative_quats)])
    self.free_dofs = None  # property used by pbd joints

  def apply(self, qp: QP) -> P:
    """Returns impulses to constrain and align bodies connected by a joint.

    Only called by springy joints.

    Args:
      qp: State data for system

    Returns:
      dP: Impulses on all bodies to maintain joint constraints
    """
    qp_p = jp.take(qp, self.body_p.idx)
    qp_c = jp.take(qp, self.body_c.idx)
    dp_p, dp_c = jp.vmap(type(self).apply_reduced)(self, qp_p, qp_c)

    # sum together all impulse contributions across parents and children
    body_idx = jp.concatenate((self.body_p.idx, self.body_c.idx))
    dp_vel = jp.concatenate((dp_p.vel, dp_c.vel))  # pytype: disable=attribute-error
    dp_ang = jp.concatenate((dp_p.ang, dp_c.ang))  # pytype: disable=attribute-error
    dp_vel = jp.segment_sum(dp_vel, body_idx, qp.pos.shape[0])
    dp_ang = jp.segment_sum(dp_ang, body_idx, qp.pos.shape[0])

    return P(vel=dp_vel, ang=dp_ang)


@pytree.register
class Revolute(Joint):
  """A springy revolute joint constrains two bodies around a single axis.

  Constructs a revolute joint where the parent's local x-axis is constrained
  to point in the same direction as the child's local x-axis.  This construction
  follows the line of nodes convention shared by the universal and spherical
  joints for x-y'-z'' intrinsic euler angles.
  """

  def __init__(self, joints: List[config_pb2.Joint], body: bodies.Body):
    super().__init__(joints, body, .5)

  def apply_reduced(self, qp_p: QP, qp_c: QP) -> Tuple[P, P]:
    """Returns calculated impulses in compressed joint space."""
    pos_p, vel_p = qp_p.to_world(self.off_p)
    pos_c, vel_c = qp_c.to_world(self.off_c)

    # push the bodies towards their offsets
    # TODO: cap this damping so that it can't overcorrect
    impulse = (pos_p - pos_c) * self.stiffness + self.spring_damping * (
        vel_p - vel_c)
    dp_p = self.body_p.impulse(qp_p, -impulse, pos_p)
    dp_c = self.body_c.impulse(qp_c, impulse, pos_c)

    # torque the bodies to align their axes
    (axis,), (angle,) = self.axis_angle(qp_p, qp_c)
    axis_c = math.rotate(self.axis_c[0], qp_c.rot)
    torque = self.stiffness * jp.cross(axis, axis_c)

    # torque the bodies to stay within angle limits
    dang = jp.where(angle < self.limit[0][0], self.limit[0][0] - angle, 0)
    dang = jp.where(angle > self.limit[0][1], self.limit[0][1] - angle, dang)
    torque -= self.limit_strength * axis * dang

    # damp the angular motion
    torque -= self.angular_damping * (qp_p.ang - qp_c.ang)

    dang_p = self.body_p.inertia * torque
    dang_c = -self.body_c.inertia * torque
    dp_p = dp_p.replace(ang=dp_p.ang + dang_p)
    dp_c = dp_c.replace(ang=dp_c.ang + dang_c)

    return dp_p, dp_c

  def axis_angle(self, qp_p: QP, qp_c: QP) -> Tuple[Any, Any]:
    """Returns axes and angles of a single joint."""
    axis_p = math.rotate(self.axis_p[0], qp_p.rot)
    ref_p = math.rotate(self.axis_p[2], qp_p.rot)
    ref_c = math.rotate(self.axis_c[2], qp_c.rot)
    # algebraically the same as the calculation in `Spherical`, but simpler
    # because child local-x and parent local-x are constrained to be the same
    psi = math.signed_angle(axis_p, ref_p, ref_c)
    return (axis_p,), (psi,)


@pytree.register
class Universal(Joint):
  """A springy universal joint constrains two bodies around two axes.

  Constructs a universal joint defined as the first two degrees of freedom
  of a spherical joint.  See `SpringSpherical` for details.
  """

  def apply_reduced(self, qp_p: QP, qp_c: QP) -> Tuple[P, P]:
    """Returns calculated impulses in compressed joint space."""
    pos_p, vel_p = qp_p.to_world(self.off_p)
    pos_c, vel_c = qp_c.to_world(self.off_c)

    # push the bodies towards their offsets
    # TODO: cap this damping so that it can't overcorrect
    impulse = (pos_p - pos_c) * self.stiffness + self.spring_damping * (
        vel_p - vel_c)
    dp_p = self.body_p.impulse(qp_p, -impulse, pos_p)
    dp_c = self.body_c.impulse(qp_c, impulse, pos_c)

    # torque the bodies to align to a joint plane
    (axis_1, axis_2), angles = self.axis_angle(qp_p, qp_c)
    axis_c_proj = axis_2 - jp.dot(axis_2, axis_1) * axis_1
    axis_c_proj = axis_c_proj / jp.safe_norm(axis_c_proj)
    torque = (self.limit_strength / 5.) * jp.cross(axis_c_proj, axis_2)

    # torque the bodies to stay within angle limits
    axis, angle = jp.array((axis_1, axis_2)), jp.array(angles)
    dang = jp.where(angle < self.limit[:, 0], self.limit[:, 0] - angle, 0)
    dang = jp.where(angle > self.limit[:, 1], self.limit[:, 1] - angle, dang)
    torque -= self.limit_strength * jp.sum(jp.vmap(jp.multiply)(axis, dang), 0)

    # damp the angular motion
    torque -= self.angular_damping * (qp_p.ang - qp_c.ang)

    dang_p = self.body_p.inertia * torque
    dang_c = -self.body_c.inertia * torque
    dp_p = dp_p.replace(ang=dp_p.ang + dang_p)
    dp_c = dp_c.replace(ang=dp_c.ang + dang_c)

    return dp_p, dp_c

  def axis_angle(self, qp_p: QP, qp_c: QP) -> Tuple[Any, Any]:
    """Returns axes and angles of a single joint."""
    v_rot = jp.vmap(math.rotate, include=[True, False])
    axis_p_rotated = v_rot(self.axis_p, qp_p.rot)
    axis_c_rotated = v_rot(self.axis_c, qp_c.rot)
    axis_1_p = axis_p_rotated[0]
    axis_2_p = axis_p_rotated[1]
    axis_1_c = axis_c_rotated[0]
    axis_2_c = axis_c_rotated[1]
    axis_3_c = axis_c_rotated[2]

    line_of_nodes = jp.cross(axis_3_c, axis_1_p)
    line_of_nodes = line_of_nodes / (1e-10 + jp.safe_norm(line_of_nodes))
    y_n_normal = axis_1_p
    psi = math.signed_angle(y_n_normal, axis_2_p, line_of_nodes)
    axis_1_p_in_xz_c = jp.dot(axis_1_p, axis_1_c) * axis_1_c + jp.dot(
        axis_1_p, axis_2_c) * axis_2_c
    axis_1_p_in_xz_c = axis_1_p_in_xz_c / (1e-10 +
                                           jp.safe_norm(axis_1_p_in_xz_c))
    theta = jp.safe_arccos(jp.clip(jp.dot(axis_1_p_in_xz_c, axis_1_p), -1,
                                   1)) * jp.sign(jp.dot(axis_1_p, axis_3_c))
    axis = (axis_1_p, axis_2_c)
    angle = (psi, theta)

    return axis, angle


@pytree.register
class Spherical(Joint):
  """A springy spherical joint constrains two bodies around three axes.

  Constructs a spherical joint which returns intrinsic euler angles in the
    x-y'-z'' convention between the parent and child.  Uses the line of nodes
    construction described in section 3.2.3.2 here:
    https://www.sedris.org/wg8home/Documents/WG80485.pdf
  """

  def apply_reduced(self, qp_p: QP, qp_c: QP) -> Tuple[P, P]:
    """Returns calculated impulses in compressed joint space."""
    pos_p, vel_p = qp_p.to_world(self.off_p)
    pos_c, vel_c = qp_c.to_world(self.off_c)

    # push the bodies towards their offsets
    # TODO: cap this damping so that it can't overcorrect
    impulse = (pos_p - pos_c) * self.stiffness + self.spring_damping * (
        vel_p - vel_c)
    dp_p = self.body_p.impulse(qp_p, -impulse, pos_p)
    dp_c = self.body_c.impulse(qp_c, impulse, pos_c)

    # torque the bodies to stay within angle limits
    axes, angles = self.axis_angle(qp_p, qp_c)
    axis, angle = jp.array(axes), jp.array(angles)
    dang = jp.where(angle < self.limit[:, 0], self.limit[:, 0] - angle, 0)
    dang = jp.where(angle > self.limit[:, 1], self.limit[:, 1] - angle, dang)
    torque = -self.limit_strength * jp.sum(jp.vmap(jp.multiply)(axis, dang), 0)

    # damp the angular motion
    torque -= self.angular_damping * (qp_p.ang - qp_c.ang)

    dp_p = dp_p.replace(ang=dp_p.ang + self.body_p.inertia * torque)
    dp_c = dp_c.replace(ang=dp_c.ang + -self.body_c.inertia * torque)

    return dp_p, dp_c

  def axis_angle(self, qp_p: QP, qp_c: QP) -> Tuple[Any, Any]:
    """Returns axes and angles of a single joint."""
    v_rot = jp.vmap(math.rotate, include=[True, False])
    axis_p_rotated = v_rot(self.axis_p, qp_p.rot)
    axis_c_rotated = v_rot(self.axis_c, qp_c.rot)
    axis_1_p = axis_p_rotated[0]
    axis_2_p = axis_p_rotated[1]
    axis_1_c = axis_c_rotated[0]
    axis_2_c = axis_c_rotated[1]
    axis_3_c = axis_c_rotated[2]

    line_of_nodes = jp.cross(axis_3_c, axis_1_p)
    line_of_nodes = line_of_nodes / (1e-10 + jp.safe_norm(line_of_nodes))
    y_n_normal = axis_1_p
    psi = math.signed_angle(y_n_normal, axis_2_p, line_of_nodes)
    axis_1_p_in_xz_c = jp.dot(axis_1_p, axis_1_c) * axis_1_c + jp.dot(
        axis_1_p, axis_2_c) * axis_2_c
    axis_1_p_in_xz_c = axis_1_p_in_xz_c / (1e-10 +
                                           jp.safe_norm(axis_1_p_in_xz_c))
    ang_between_1_p_xz_c = jp.dot(axis_1_p_in_xz_c, axis_1_p)
    theta = jp.safe_arccos(jp.clip(ang_between_1_p_xz_c, -1, 1)) * jp.sign(
        jp.dot(axis_1_p, axis_3_c))
    yc_n_normal = -axis_3_c
    phi = math.signed_angle(yc_n_normal, axis_2_c, line_of_nodes)
    axis = (axis_1_p, axis_2_c, axis_3_c)
    angle = (psi, theta, phi)

    return axis, angle


def get(config: config_pb2.Config, body: bodies.Body) -> List[Joint]:
  """Creates all joints given a config."""
  joints = {}
  for joint in config.joints:
    dof = len(joint.angle_limit)
    springy = joint.stiffness > 0
    if springy:
      if dof not in joints:
        joints[dof] = []
      joints[dof].append(joint)

  # ensure stable order for joint application: dof
  joints = sorted(joints.items(), key=lambda kv: kv[0])
  ret = []

  for dof, v in joints:
    if dof == 1:
      ret.append(Revolute(v, body))
    elif dof == 2:
      ret.append(Universal(v, body))
    elif dof == 3:
      ret.append(Spherical(v, body))
    else:
      raise RuntimeError(f'invalid number of joint limits: {dof}')

  return ret



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""D-min Sweep."""
# pylint:disable=g-complex-comprehension
from brax.experimental.braxlines.experiments import defaults

ENV_NAMES = ('ant', 'halfcheetah', 'humanoid', 'hopper', 'walker2d')
AGENT_MODULE = 'brax.experimental.braxlines.irl_smm.train'
CONFIG = [
    dict(
        env_name=[env_name],
        obs_indices='vel',
        target_num_modes=2,
        obs_scale=8,
        reward_type=['gail', 'fairl', 'gail2', 'mle', 'airl'],
        seed=list(range(10)),
        normalize_obs_for_disc=False,
        evaluate_dist=True,
        env_reward_multiplier=0.0,
        spectral_norm=False,
        gradient_penalty_weight=0.0,
        ppo_params=defaults.get_ppo_params(env_name, 2))
    for env_name in ENV_NAMES
]



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint:disable=g-multiple-import
"""Tests for spring physics pipeline."""

from absl.testing import absltest
from brax.v2 import kinematics
from brax.v2 import test_utils
from brax.v2.generalized import pipeline as g_pipeline
from brax.v2.spring import pipeline
import jax
from jax import numpy as jp
import numpy as np


class PipelineTest(absltest.TestCase):

  def test_pendulum(self):
    sys = test_utils.load_fixture('triple_pendulum.xml')
    sys = sys.replace(dt=0.0001)
    sys = sys.replace(
        link=sys.link.replace(constraint_stiffness=jp.array([800_000.0] * 3))
    )

    state = pipeline.init(sys, sys.init_q, jp.zeros(sys.qd_size()))
    j_spring_step = jax.jit(pipeline.step)
    for _ in range(10_000):
      state = j_spring_step(sys, state, jp.zeros(sys.qd_size()))
    x = state.x

    # compare against generalized step
    q, qd = jp.zeros(sys.q_size()), jp.zeros(sys.qd_size())
    state = g_pipeline.init(sys, q, qd)
    j_g_step = jax.jit(g_pipeline.step)
    j_forward = jax.jit(kinematics.forward)
    for _ in range(10_000):
      state = j_g_step(sys, state, jp.zeros(sys.qd_size()))
    x_g, _ = j_forward(sys, state.q, state.qd)

    # trajectories should be close after 1 second of simulation
    self.assertLess(jp.linalg.norm(x_g.pos - x.pos), 2e-2)

  def test_spherical_pendulum(self):
    sys = test_utils.load_fixture('single_spherical_pendulum.xml')

    init_q = jp.array([0.1, 0.2, 0.3])
    init_qd = jax.random.uniform(jax.random.PRNGKey(0), (3,)) / 10.0

    # set sensible springy defaults
    link = sys.link.replace(constraint_limit_stiffness=jp.array([0.0] * 3))
    link = link.replace(constraint_stiffness=jp.array([100_000.0] * 3))
    link = link.replace(constraint_ang_damping=jp.array([0.0] * 3))
    link = link.replace(constraint_damping=jp.array([200.0] * 3))
    sys = sys.replace(link=link)
    sys = sys.replace(ang_damping=0.0)
    sys = sys.replace(dt=0.001)
    sys = sys.replace(solver_iterations=500)

    state = pipeline.init(sys, init_q, init_qd)
    j_spring_step = jax.jit(pipeline.step)
    for _ in range(1000):
      state = j_spring_step(sys, state, jp.zeros(sys.qd_size()))
    x = state.x

    # compare against generalized step
    q, qd = init_q, init_qd
    state = g_pipeline.init(sys, q, qd)
    j_g_step = jax.jit(g_pipeline.step)
    j_forward = jax.jit(kinematics.forward)
    for _ in range(1000):
      state = j_g_step(sys, state, jp.zeros(sys.qd_size()))
    x_g, _ = j_forward(sys, state.q, state.qd)

    # trajectories should be close after 1 second of simulation
    self.assertLess(jp.linalg.norm(x_g.rot - x.rot), 1e-2)

  def test_prismatic_joint(self):
    # tests launching a mass along a slide joint in the [1,1,1] direction
    sys = test_utils.load_fixture('single_prismatic.xml')

    init_q = jp.array([0.9])
    init_qd = jp.array([1.0])

    # set sensible springy defaults
    link = sys.link.replace(constraint_limit_stiffness=jp.array([10_000.0]))
    link = link.replace(constraint_stiffness=jp.array([10_000.0]))
    link = link.replace(constraint_ang_damping=jp.array([0.0]))
    link = link.replace(constraint_damping=jp.array([200.0]))
    sys = sys.replace(link=link)
    sys = sys.replace(ang_damping=0.0)
    sys = sys.replace(dt=0.001)
    sys = sys.replace(solver_iterations=500)

    state = pipeline.init(sys, init_q, init_qd)
    j_spring_step = jax.jit(pipeline.step)
    for _ in range(1000):
      state = j_spring_step(sys, state, jp.zeros(sys.qd_size()))
    x = state.x

    # compare against generalized step
    state = g_pipeline.init(sys, init_q, init_qd)
    j_g_step = jax.jit(g_pipeline.step)
    j_forward = jax.jit(kinematics.forward)
    for _ in range(1000):
      state = j_g_step(sys, state, jp.zeros(sys.qd_size()))
    x_g, _ = j_forward(sys, state.q, state.qd)

    # trajectories should be close after 1 second of simulation
    np.testing.assert_allclose(x_g.pos, x.pos, rtol=5e-3, atol=5e-3)

  def test_sliding_capsule(self):
    sys = test_utils.load_fixture('capsule.xml')
    sys = sys.replace(dt=0.001)

    qd = jp.zeros(sys.qd_size())
    qd = qd.at[0].set(5.0)

    state = pipeline.init(sys, sys.init_q, qd)
    j_spring_step = jax.jit(pipeline.step)
    for _ in range(1000):
      state = j_spring_step(sys, state, jp.zeros(sys.qd_size()))
    x, xd = state.x, state.xd

    # capsule slides to a stop
    self.assertAlmostEqual(x.pos[0, 2], 0.25, delta=1e-3)
    np.testing.assert_allclose(
        x.rot, jp.array([[1.0, 0.0, 0.0, 0.0]]), atol=1e-3
    )
    np.testing.assert_allclose(xd.vel, jp.zeros_like(xd.vel), atol=1e-3)
    np.testing.assert_allclose(xd.ang, jp.zeros_like(xd.ang), atol=1e-3)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trains a 2D walker to run in the +x direction."""

from typing import Tuple

import brax
from brax import jumpy as jp
from brax import math
from brax.envs import env


class Walker2d(env.Env):



  """
  ### Description

  This environment builds on the hopper environment based on the work done by
  Erez, Tassa, and Todorov in
  ["Infinite Horizon Model Predictive Control for Nonlinear Periodic Tasks"](http://www.roboticsproceedings.org/rss07/p10.pdf)
  by adding another set of legs making it possible for the robot to walker
  forward instead of hop. Like other Mujoco environments, this environment aims
  to increase the number of independent state and control variables as compared
  to the classic control environments.

  The walker is a two-dimensional two-legged figure that consist of four main
  body parts - a single torso at the top (with the two legs splitting after the
  torso), two thighs in the middle below the torso, two legs in the bottom below
  the thighs, and two feet attached to the legs on which the entire body rests.

  The goal is to make coordinate both sets of feet, legs, and thighs to move in
  the forward (right) direction by applying torques on the six hinges connecting
  the six body parts.

  ### Action Space

  The agent take a 6-element vector for actions. The action space is a
  continuous `(action, action, action, action, action, action)` all in
  `[-1, 1]`, where `action` represents the numerical torques applied at the
  hinge joints.

  | Num | Action                                 | Control Min | Control Max | Name (in corresponding config) | Joint | Unit         |
  |-----|----------------------------------------|-------------|-------------|--------------------------------|-------|--------------|
  | 0   | Torque applied on the thigh rotor      | -1          | 1           | thigh_joint                    | hinge | torque (N m) |
  | 1   | Torque applied on the leg rotor        | -1          | 1           | leg_joint                      | hinge | torque (N m) |
  | 2   | Torque applied on the foot rotor       | -1          | 1           | foot_joint                     | hinge | torque (N m) |
  | 3   | Torque applied on the left thigh rotor | -1          | 1           | thigh_left_joint               | hinge | torque (N m) |
  | 4   | Torque applied on the left leg rotor   | -1          | 1           | leg_left_joint                 | hinge | torque (N m) |
  | 5   | Torque applied on the left foot rotor  | -1          | 1           | foot_left_joint                | hinge | torque (N m) |

  ### Observation Space

  The state space consists of positional values of different body parts of the
  walker, followed by the velocities of those individual parts (their
  derivatives) with all the positions ordered before all the velocities.

  The observation is a `ndarray` with shape `(17,)` where the elements
  correspond to the following:

  | Num | Observation                                      | Min  | Max | Name (in corresponding config) | Joint | Unit                     |
  |-----|--------------------------------------------------|------|-----|--------------------------------|-------|--------------------------|
  | 0   | z-coordinate of the top (height of hopper)       | -Inf | Inf | rootz (torso)                  | slide | position (m)             |
  | 1   | angle of the top                                 | -Inf | Inf | rooty (torso)                  | hinge | angle (rad)              |
  | 2   | angle of the thigh joint                         | -Inf | Inf | thigh_joint                    | hinge | angle (rad)              |
  | 3   | angle of the leg joint                           | -Inf | Inf | leg_joint                      | hinge | angle (rad)              |
  | 4   | angle of the foot joint                          | -Inf | Inf | foot_joint                     | hinge | angle (rad)              |
  | 5   | angle of the left thigh joint                    | -Inf | Inf | thigh_left_joint               | hinge | angle (rad)              |
  | 6   | angle of the left leg joint                      | -Inf | Inf | leg_left_joint                 | hinge | angle (rad)              |
  | 7   | angle of the left foot joint                     | -Inf | Inf | foot_left_joint                | hinge | angle (rad)              |
  | 8   | velocity of the x-coordinate of the top          | -Inf | Inf | rootx                          | slide | velocity (m/s)           |
  | 9   | velocity of the z-coordinate (height) of the top | -Inf | Inf | rootz                          | slide | velocity (m/s)           |
  | 10  | angular velocity of the angle of the top         | -Inf | Inf | rooty                          | hinge | angular velocity (rad/s) |
  | 11  | angular velocity of the thigh hinge              | -Inf | Inf | thigh_joint                    | hinge | angular velocity (rad/s) |
  | 12  | angular velocity of the leg hinge                | -Inf | Inf | leg_joint                      | hinge | angular velocity (rad/s) |
  | 13  | angular velocity of the foot hinge               | -Inf | Inf | foot_joint                     | hinge | angular velocity (rad/s) |
  | 14  | angular velocity of the thigh hinge              | -Inf | Inf | thigh_left_joint               | hinge | angular velocity (rad/s) |
  | 15  | angular velocity of the leg hinge                | -Inf | Inf | leg_left_joint                 | hinge | angular velocity (rad/s) |
  | 16  | angular velocity of the foot hinge               | -Inf | Inf | foot_left_joint                | hinge | angular velocity (rad/s) |

  ### Rewards

  The reward consists of three parts:
  - *reward_healthy*: Every timestep that the walker is alive, it gets a reward of
    1
  - *reward_forward*: A reward of walking forward which is measured as
    *(x-coordinate before action - x-coordinate after action) / dt*.  *dt* is
    the time between actions - the default *dt = 0.008*. This reward would be
    positive if the walker walks forward (right) desired.
  - *reward_ctrl*: A negative reward for penalising the walker if it takes
    actions that are too large. It is measured as *-coefficient **x**
    sum(action<sup>2</sup>)* where *coefficient* is a parameter set for the
    control and has a default value of 0.001

  ### Starting State

  All observations start in state (0.0, 1.25, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
  0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0) with a uniform noise in the range of
  [-0.005, 0.005] added to the values for stochasticity.

  ### Episode Termination

  The episode terminates when any of the following happens:

  1. The episode duration reaches a 1000 timesteps
  2. The height of the walker is ***not*** in the range `[0.7, 2]`
  3. The absolute value of the angle is ***not*** in the range `[-1, 1]`

  ### Arguments

  No additional arguments are currently supported (in v2 and lower), but
  modifications can be made to the XML file in the assets folder (or by changing
  the path to a modified XML file in another folder)..

  ```
  env = gym.make('Walker2d-v2')
  ```

  v3, v4 and v5 take gym.make kwargs such as ctrl_cost_weight, reset_noise_scale
  etc.

  ```
  env = gym.make('Walker2d-v5', ctrl_cost_weight=0.1, ....)
  ```

  ### Version History

  * v5: ported to Brax.
  * v4: all mujoco environments now use the mujoco bindings in mujoco>=2.1.3
  * v3: support for gym.make kwargs such as xml_file, ctrl_cost_weight,
    reset_noise_scale etc. rgb rendering comes from tracking camera (so agent
    does not run away from screen)
  * v2: All continuous control environments now use mujoco_py >= 1.50
  * v1: max_time_steps raised to 1000 for robot based tasks. Added
    reward_threshold to environments.
  * v0: Initial versions release (1.0.0)
  """


  def __init__(self,
               forward_reward_weight: float = 1.0,
               ctrl_cost_weight: float = 1e-3,
               healthy_reward: float = 1.0,
               terminate_when_unhealthy: bool = True,
               healthy_z_range: Tuple[float, float] = (0.7, 2.0),
               healthy_angle_range=(-1.0, 1.0),
               reset_noise_scale=5e-3,
               exclude_current_positions_from_observation=True,
               legacy_spring=False,
               **kwargs):
    """Creates a Walker environment.

    Args:
      forward_reward_weight: Weight for the forward reward, i.e. velocity in
        x-direction.
      ctrl_cost_weight: Weight for the control cost.
      healthy_reward: Reward for staying healthy, i.e. respecting the posture
        constraints.
      terminate_when_unhealthy: Done bit will be set when unhealthy if true.
      healthy_z_range: Range of the z-position for being healthy.
      healthy_angle_range: Range of joint angles for being healthy.
      reset_noise_scale: Scale of noise to add to reset states.
      exclude_current_positions_from_observation: x-position will be hidden
        from the observations if true.
      legacy_spring: if True, reverts to legacy spring dynamics instead of pbd.
      **kwargs: Arguments that are passed to the base class.
    """
    config = _SYSTEM_CONFIG_SPRING if legacy_spring else _SYSTEM_CONFIG
    super().__init__(config=config, **kwargs)

    self._forward_reward_weight = forward_reward_weight
    self._ctrl_cost_weight = ctrl_cost_weight
    self._healthy_reward = healthy_reward
    self._terminate_when_unhealthy = terminate_when_unhealthy
    self._healthy_z_range = healthy_z_range
    self._healthy_angle_range = healthy_angle_range
    self._reset_noise_scale = reset_noise_scale
    self._exclude_current_positions_from_observation = (
        exclude_current_positions_from_observation
    )

  def reset(self, rng: jp.ndarray) -> env.State:
    """Resets the environment to an initial state."""
    rng, rng1, rng2 = jp.random_split(rng, 3)
    qpos = self.sys.default_angle() + self._noise(rng1)
    qvel = self._noise(rng2)
    qp = self.sys.default_qp(joint_angle=qpos, joint_velocity=qvel)
    obs = self._get_obs(qp)
    reward, done, zero = jp.zeros(3)
    metrics = {
        'reward_forward': zero,
        'reward_ctrl': zero,
        'reward_healthy': zero,
        'x_position': zero,
        'x_velocity': zero,
    }
    return env.State(qp, obs, reward, done, metrics)

  def step(self, state: env.State, action: jp.ndarray) -> env.State:
    """Run one timestep of the environment's dynamics."""
    qp, _ = self.sys.step(state.qp, action)

    x_velocity = (qp.pos[0, 0] - state.qp.pos[0, 0]) / self.sys.config.dt
    forward_reward = self._forward_reward_weight * x_velocity

    min_z, max_z = self._healthy_z_range
    min_angle, max_angle = self._healthy_angle_range
    ang_y = math.quat_to_euler(qp.rot[0])[1]
    is_healthy = jp.where(qp.pos[0, 2] < min_z, x=0.0, y=1.0)
    is_healthy = jp.where(qp.pos[0, 2] > max_z, x=0.0, y=is_healthy)
    is_healthy = jp.where(ang_y > max_angle, x=0.0, y=is_healthy)
    is_healthy = jp.where(ang_y < min_angle, x=0.0, y=is_healthy)
    if self._terminate_when_unhealthy:
      healthy_reward = self._healthy_reward
    else:
      healthy_reward = self._healthy_reward * is_healthy

    ctrl_cost = self._ctrl_cost_weight * jp.sum(jp.square(action))

    obs = self._get_obs(qp)
    reward = forward_reward + healthy_reward - ctrl_cost
    done = 1.0 - is_healthy if self._terminate_when_unhealthy else 0.0
    state.metrics.update(
        reward_forward=forward_reward,
        reward_ctrl=-ctrl_cost,
        reward_healthy=healthy_reward,
        x_position=qp.pos[0, 0],
        x_velocity=x_velocity)

    return state.replace(qp=qp, obs=obs, reward=reward, done=done)

  def _get_obs(self, qp: brax.QP) -> jp.ndarray:
    """Returns the environment observations."""
    joint_angle, joint_vel = self.sys.joints[0].angle_vel(qp)
    ang_y = math.quat_to_euler(qp.rot[0])[1:2]

    # qpos: position and orientation of the torso and the joint angles
    if self._exclude_current_positions_from_observation:
      qpos = [qp.pos[0, 2:], ang_y, joint_angle]
    else:
      qpos = [qp.pos[0, (0, 2)], ang_y, joint_angle]

    # qvel: velocity of the torso and the joint angle velocities
    qvel = [qp.vel[0, (0, 2)], qp.ang[0, 1:2], joint_vel]

    return jp.concatenate(qpos + qvel)

  def _noise(self, rng):
    low, hi = -self._reset_noise_scale, self._reset_noise_scale
    return jp.random_uniform(rng, (self.sys.num_joint_dof,), low, hi)


_SYSTEM_CONFIG = """
  bodies {
    name: "torso"
    colliders {
      capsule {
        radius: 0.05
        length: 0.5
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 3.6651914
  }
  bodies {
    name: "thigh"
    colliders {
      position { z: -0.225 }
      rotation {}
      capsule {
        radius: 0.05
        length: 0.55
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 4.0578904
  }
  bodies {
    name: "leg"
    colliders {
      position {}
      rotation {}
      capsule {
        radius: 0.04
        length: 0.58
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 2.7813568
  }
  bodies {
    name: "foot"
    colliders {
      position {
        x: -0.1
        y: -0.2
        z: -0.1
      }
      rotation { y: 90.0 }
      capsule {
        radius: 0.06
        length: 0.32
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 3.1667254
  }
  bodies {
    name: "thigh_left"
    colliders {
      position { z: -0.225 }
      rotation {}
      capsule {
        radius: 0.05
        length: 0.55
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 4.0578904
  }
  bodies {
    name: "leg_left"
    colliders {
      position {}
      rotation {}
      capsule {
        radius: 0.04
        length: 0.58
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 2.7813568
  }
  bodies {
    name: "foot_left"
    colliders {
      position {
        x: -0.1
        y: -0.2
        z: -0.1
      }
      rotation { y: 90.0 }
      capsule {
        radius: 0.06
        length: 0.32
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 3.1667254
  }
  bodies {
    name: "floor"
    colliders {
      plane {}
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    frozen { all: true }
  }
  joints {
    name: "thigh_joint"
    parent: "torso"
    child: "thigh"
    parent_offset { z: -0.2 }
    rotation { z: -90.0 }
    angle_limit { min: -150.0 }
  }
  joints {
    name: "leg_joint"
    parent: "thigh"
    child: "leg"
    parent_offset { z: -0.45 }
    child_offset { z: 0.25 }
    rotation { z: -90.0 }
    angle_limit { min: -150.0 }
  }
  joints {
    name: "foot_joint"
    parent: "leg"
    child: "foot"
    parent_offset { z: -0.25 }
    child_offset {
      x: -0.2
      y: -0.2
      z: -0.1
    }
    rotation { z: -90.0 }
    angle_limit { min: -45.0 max: 45.0 }
  }
  joints {
    name: "thigh_left_joint"
    parent: "torso"
    child: "thigh_left"
    parent_offset { z: -0.2 }
    rotation { z: -90.0 }
    angle_limit { min: -150.0 }
  }
  joints {
    name: "leg_left_joint"
    parent: "thigh_left"
    child: "leg_left"
    parent_offset { z: -0.45 }
    child_offset { z: 0.25 }
    rotation { z: -90.0 }
    angle_limit { min: -150.0 }
  }
  joints {
    name: "foot_left_joint"
    parent: "leg_left"
    child: "foot_left"
    parent_offset { z: -0.25 }
    child_offset {
      x: -0.2
      y: -0.2
      z: -0.1
    }
    rotation { z: -90.0 }
    angle_limit { min: -45.0 max: 45.0 }
  }
  actuators {
    name: "thigh_joint"
    joint: "thigh_joint"
    strength: 100.0
    torque {}
  }
  actuators {
    name: "leg_joint"
    joint: "leg_joint"
    strength: 100.0
    torque {}
  }
  actuators {
    name: "foot_joint"
    joint: "foot_joint"
    strength: 100.0
    torque {}
  }
  actuators {
    name: "thigh_left_joint"
    joint: "thigh_left_joint"
    strength: 100.0
    torque {}
  }
  actuators {
    name: "leg_left_joint"
    joint: "leg_left_joint"
    strength: 100.0
    torque {}
  }
  actuators {
    name: "foot_left_joint"
    joint: "foot_left_joint"
    strength: 100.0
    torque {}
  }
  friction: 0.94868329805
  gravity { z: -9.81 }
  velocity_damping: 1.0
  angular_damping: -0.05
  collide_include {
    first: "floor"
    second: "torso"
  }
  collide_include {
    first: "floor"
    second: "thigh"
  }
  collide_include {
    first: "floor"
    second: "leg"
  }
  collide_include {
    first: "floor"
    second: "foot"
  }
  collide_include {
    first: "floor"
    second: "thigh_left"
  }
  collide_include {
    first: "floor"
    second: "leg_left"
  }
  collide_include {
    first: "floor"
    second: "foot_left"
  }
  dt: 0.008
  substeps: 4
  frozen {
    position { y: 1.0 }
    rotation { x: 1.0 z: 1.0 }
  }
  defaults {
    qps { name: "torso" pos { z: 1.19 } }
    angles { name: "thigh_joint" angle {} }
    angles { name: "leg_joint" angle {} }
    angles { name: "thigh_left_joint" angle {} }
    angles { name: "leg_left_joint" angle {} }
  }
  dynamics_mode: "pbd"
  """

_SYSTEM_CONFIG_SPRING = """
  bodies {
    name: "torso"
    colliders {
      capsule {
        radius: 0.05
        length: 0.5
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 3.6651914
  }
  bodies {
    name: "thigh"
    colliders {
      position { z: -0.225 }
      rotation {}
      capsule {
        radius: 0.05
        length: 0.55
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 4.0578904
  }
  bodies {
    name: "leg"
    colliders {
      position {}
      rotation {}
      capsule {
        radius: 0.04
        length: 0.58
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 2.7813568
  }
  bodies {
    name: "foot"
    colliders {
      position {
        x: -0.1
        y: -0.2
        z: -0.1
      }
      rotation { y: 90.0 }
      capsule {
        radius: 0.06
        length: 0.32
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 3.1667254
  }
  bodies {
    name: "thigh_left"
    colliders {
      position { z: -0.225 }
      rotation {}
      capsule {
        radius: 0.05
        length: 0.55
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 4.0578904
  }
  bodies {
    name: "leg_left"
    colliders {
      position {}
      rotation {}
      capsule {
        radius: 0.04
        length: 0.58
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 2.7813568
  }
  bodies {
    name: "foot_left"
    colliders {
      position {
        x: -0.1
        y: -0.2
        z: -0.1
      }
      rotation { y: 90.0 }
      capsule {
        radius: 0.06
        length: 0.32
      }
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    mass: 3.1667254
  }
  bodies {
    name: "floor"
    colliders {
      plane {}
    }
    inertia { x: 1.0 y: 1.0 z: 1.0 }
    frozen { all: true }
  }
  joints {
    name: "thigh_joint"
    stiffness: 10000.0
    parent: "torso"
    child: "thigh"
    parent_offset { z: -0.2 }
    rotation { z: -90.0 }
    angle_limit { min: -150.0 }
    angular_damping: 20.0
  }
  joints {
    name: "leg_joint"
    stiffness: 10000.0
    parent: "thigh"
    child: "leg"
    parent_offset { z: -0.45 }
    child_offset { z: 0.25 }
    rotation { z: -90.0 }
    angle_limit { min: -150.0 }
    angular_damping: 20.0
  }
  joints {
    name: "foot_joint"
    stiffness: 10000.0
    parent: "leg"
    child: "foot"
    parent_offset { z: -0.25 }
    child_offset {
      x: -0.2
      y: -0.2
      z: -0.1
    }
    rotation { z: -90.0 }
    angle_limit { min: -45.0 max: 45.0 }
    angular_damping: 20.0
  }
  joints {
    name: "thigh_left_joint"
    stiffness: 10000.0
    parent: "torso"
    child: "thigh_left"
    parent_offset { z: -0.2 }
    rotation { z: -90.0 }
    angle_limit { min: -150.0 }
    angular_damping: 20.0
  }
  joints {
    name: "leg_left_joint"
    stiffness: 10000.0
    parent: "thigh_left"
    child: "leg_left"
    parent_offset { z: -0.45 }
    child_offset { z: 0.25 }
    rotation { z: -90.0 }
    angle_limit { min: -150.0 }
    angular_damping: 20.0
  }
  joints {
    name: "foot_left_joint"
    stiffness: 10000.0
    parent: "leg_left"
    child: "foot_left"
    parent_offset { z: -0.25 }
    child_offset {
      x: -0.2
      y: -0.2
      z: -0.1
    }
    rotation { z: -90.0 }
    angle_limit { min: -45.0 max: 45.0 }
    angular_damping: 20.0
  }
  actuators {
    name: "thigh_joint"
    joint: "thigh_joint"
    strength: 100.0
    torque {}
  }
  actuators {
    name: "leg_joint"
    joint: "leg_joint"
    strength: 100.0
    torque {}
  }
  actuators {
    name: "foot_joint"
    joint: "foot_joint"
    strength: 100.0
    torque {}
  }
  actuators {
    name: "thigh_left_joint"
    joint: "thigh_left_joint"
    strength: 100.0
    torque {}
  }
  actuators {
    name: "leg_left_joint"
    joint: "leg_left_joint"
    strength: 100.0
    torque {}
  }
  actuators {
    name: "foot_left_joint"
    joint: "foot_left_joint"
    strength: 100.0
    torque {}
  }
  friction: 0.94868329805
  gravity { z: -9.81 }
  velocity_damping: 1.0
  angular_damping: -0.05
  baumgarte_erp: 0.1
  collide_include {
    first: "floor"
    second: "torso"
  }
  collide_include {
    first: "floor"
    second: "thigh"
  }
  collide_include {
    first: "floor"
    second: "leg"
  }
  collide_include {
    first: "floor"
    second: "foot"
  }
  collide_include {
    first: "floor"
    second: "thigh_left"
  }
  collide_include {
    first: "floor"
    second: "leg_left"
  }
  collide_include {
    first: "floor"
    second: "foot_left"
  }
  dt: 0.008
  substeps: 4
  frozen {
    position { y: 1.0 }
    rotation { x: 1.0 z: 1.0 }
  }
  defaults {
    qps { name: "torso" pos { z: 1.19 } }
    angles { name: "thigh_joint" angle {} }
    angles { name: "leg_joint" angle {} }
    angles { name: "thigh_left_joint" angle {} }
    angles { name: "leg_left_joint" angle {} }
  }
  dynamics_mode: "legacy_spring"
  """



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for spring physics joints."""

from absl.testing import absltest
from absl.testing import parameterized
from brax.v2 import kinematics
from brax.v2 import test_utils
from brax.v2.spring import pipeline
import jax
from jax import numpy as jp


class JointTest(parameterized.TestCase):

  @parameterized.parameters(
      (2.0, 0.125, 0.0625), (5.0, 0.125, 0.03125), (1.0, 0.0625, 0.1)
  )
  def test_pendulum_period(self, mass, radius, vel):
    """A small spherical mass swings for approximately one period."""
    # config = text_format.Parse(JointTest._CONFIG, brax.Config())
    sys = test_utils.load_fixture('single_pendulum.xml')

    dist_to_anchor = 0.5
    inertia_cm = 2.0 / 5.0 * mass * radius**2.0
    inertia_about_anchor = mass * dist_to_anchor**2.0 + inertia_cm
    g = 9.81
    # formula for period of pendulum
    period = (
        2 * jp.pi * jp.sqrt(inertia_about_anchor / (mass * g * dist_to_anchor))
    )
    num_timesteps = 20_000
    sys = sys.replace(dt=period / num_timesteps)
    link = sys.link.replace(constraint_limit_stiffness=jp.array([0.0] * 1))
    link = link.replace(constraint_stiffness=jp.array([10_000.0] * 1))
    link = link.replace(constraint_ang_damping=jp.array([0.0] * 1))
    link = link.replace(constraint_damping=jp.array([0.0] * 1))
    sys = sys.replace(link=link)
    sys = sys.replace(ang_damping=0.0)
    sys = sys.replace(
        link=sys.link.replace(
            inertia=sys.link.inertia.replace(
                i=jp.array([0.4 * mass * radius**2 * jp.eye(3)] * 1),
                mass=jp.array([mass]),
            )
        )
    )

    # init with small initial velocity for small angle approx. validity
    state = pipeline.init(sys, sys.init_q, jp.zeros(sys.qd_size()))
    x, xd = kinematics.forward(
        sys,
        jp.array(
            [
                -jp.pi / 2.0,
            ]
        ),
        jp.array([vel]),
    )
    state = state.replace(x=x, xd=xd)

    j_spring_step = jax.jit(pipeline.step)
    for _ in range(num_timesteps):
      state = j_spring_step(sys, state, jp.zeros(sys.act_size()))

    self.assertAlmostEqual(state.xd.ang[0, 0], vel, 2)  # returned to the origin


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Humanoid."""
# pylint:disable=protected-access
import functools
from brax.envs import humanoid
from brax.experimental.composer import component_editor
from brax.experimental.composer.components import common


COLLIDES = ('torso', 'left_shin', 'right_shin')

ROOT = 'torso'

DEFAULT_OBSERVERS = ('root_z_joints',)

TERM_FN = functools.partial(
    common.height_term_fn, max_height=2.1, min_height=0.7)


def get_specs():
  return dict(
      message_str=component_editor.filter_message_str(
          humanoid._SYSTEM_CONFIG_SPRING, 'floor'),
      collides=COLLIDES,
      root=ROOT,
      term_fn=TERM_FN,
      observers=DEFAULT_OBSERVERS)



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Trains an agent to locomote to a target location."""

from typing import Tuple

import brax
from brax import jumpy as jp
from brax import math
from brax.envs import env


class Fetch(env.Env):
  """Fetch trains a dog to run to a target location."""

  def __init__(self, legacy_spring=False, **kwargs):
    config = _SYSTEM_CONFIG_SPRING if legacy_spring else _SYSTEM_CONFIG
    super().__init__(config=config, **kwargs)
    self.target_idx = self.sys.body.index['Target']
    self.torso_idx = self.sys.body.index['Torso']
    self.target_radius = 2
    self.target_distance = 15

  def reset(self, rng: jp.ndarray) -> env.State:
    qp = self.sys.default_qp()
    rng, target = self._random_target(rng)
    pos = jp.index_update(qp.pos, self.target_idx, target)
    qp = qp.replace(pos=pos)
    info = self.sys.info(qp)
    obs = self._get_obs(qp, info)
    reward, done, zero = jp.zeros(3)
    metrics = {
        'hits': zero,
        'weightedHits': zero,
        'movingToTarget': zero,
        'torsoIsUp': zero,
        'torsoHeight': zero
    }
    info = {'rng': rng}
    return env.State(qp, obs, reward, done, metrics, info)

  def step(self, state: env.State, action: jp.ndarray) -> env.State:
    qp, info = self.sys.step(state.qp, action)
    obs = self._get_obs(qp, info)

    # small reward for torso moving towards target
    torso_delta = qp.pos[self.torso_idx] - state.qp.pos[self.torso_idx]
    target_rel = qp.pos[self.target_idx] - qp.pos[self.torso_idx]
    target_dist = jp.norm(target_rel)
    target_dir = target_rel / (1e-6 + target_dist)
    moving_to_target = .1 * jp.dot(torso_delta, target_dir)

    # small reward for torso being up
    up = jp.array([0., 0., 1.])
    torso_up = math.rotate(up, qp.rot[self.torso_idx])
    torso_is_up = .1 * self.sys.config.dt * jp.dot(torso_up, up)

    # small reward for torso height
    torso_height = .1 * self.sys.config.dt * qp.pos[0, 2]

    # big reward for reaching target and facing it
    fwd = jp.array([1., 0., 0.])
    torso_fwd = math.rotate(fwd, qp.rot[self.torso_idx])
    torso_facing = jp.dot(target_dir, torso_fwd)
    target_hit = target_dist < self.target_radius
    target_hit = jp.where(target_hit, jp.float32(1), jp.float32(0))
    weighted_hit = target_hit * torso_facing

    reward = torso_height + moving_to_target + torso_is_up + weighted_hit

    state.metrics.update(
        hits=target_hit,
        weightedHits=weighted_hit,
        movingToTarget=moving_to_target,
        torsoIsUp=torso_is_up,
        torsoHeight=torso_height)

    # teleport any hit targets
    rng, target = self._random_target(state.info['rng'])
    target = jp.where(target_hit, target, qp.pos[self.target_idx])
    pos = jp.index_update(qp.pos, self.target_idx, target)
    qp = qp.replace(pos=pos)
    state.info.update(rng=rng)
    return state.replace(qp=qp, obs=obs, reward=reward)

  def _get_obs(self, qp: brax.QP, info: brax.Info) -> jp.ndarray:
    """Egocentric observation of target and the dog's body."""
    torso_fwd = math.rotate(jp.array([1., 0., 0.]), qp.rot[self.torso_idx])
    torso_up = math.rotate(jp.array([0., 0., 1.]), qp.rot[self.torso_idx])

    v_inv_rotate = jp.vmap(math.inv_rotate, include=(True, False))

    pos_local = qp.pos - qp.pos[self.torso_idx]
    pos_local = v_inv_rotate(pos_local, qp.rot[self.torso_idx])
    vel_local = v_inv_rotate(qp.vel, qp.rot[self.torso_idx])

    target_local = pos_local[self.target_idx]
    target_local_mag = jp.reshape(jp.norm(target_local), -1)
    target_local_dir = target_local / (1e-6 + target_local_mag)

    pos_local = jp.reshape(pos_local, -1)
    vel_local = jp.reshape(vel_local, -1)

    contact_mag = jp.sum(jp.square(info.contact.vel), axis=-1)
    contacts = jp.where(contact_mag > 0.00001, 1, 0)

    return jp.concatenate([
        torso_fwd, torso_up, target_local_mag, target_local_dir, pos_local,
        vel_local, contacts
    ])

  def _random_target(self, rng: jp.ndarray) -> Tuple[jp.ndarray, jp.ndarray]:
    """Returns a target location in a random circle on xz plane."""
    rng, rng1, rng2 = jp.random_split(rng, 3)
    dist = self.target_radius + self.target_distance * jp.random_uniform(rng1)
    ang = jp.pi * 2. * jp.random_uniform(rng2)
    target_x = dist * jp.cos(ang)
    target_y = dist * jp.sin(ang)
    target_z = 1.0
    target = jp.array([target_x, target_y, target_z]).transpose()
    return rng, target


_SYSTEM_CONFIG = """
  bodies {
    name: "Torso"
    colliders {
      box {
        halfsize { x: 0.75 y: 0.25 z: 0.125 }
      }
    }
    inertia { x: 1 y: 1 z: 1 }
    mass: 1.0
  }
  bodies {
    name: "Shoulders"
    colliders {
      box {
        halfsize { x: 0.25 y: 0.75 z: 0.125 }
      }
    }
    inertia { x: 1 y: 1 z: 1 }
    mass: 1.0
  }
  bodies {
    name: "Hips"
    colliders {
      box {
        halfsize { x: 0.25 y: 0.75 z: 0.125 }
      }
    }
    inertia { x: 1 y: 1 z: 1 }
    mass: 1.0
  }
  bodies {
    name: "Front Right Upper"
    colliders {
      box {
        halfsize { x: 0.25 y: 0.125 z: 0.5 }
      }
    }
    inertia { x: 1 y: 1 z: 1 }
    mass: 1.0
  }
  bodies {
    name: "Front Right Lower"
    colliders {
      box {
        halfsize { x: 0.25 y: 0.125 z: 0.5 }
      }
    }
    inertia { x: 1 y: 1 z: 1 }
    mass: 1.0
  }

  bodies {
    name: "Front Left Upper"
    colliders {
      box {
        halfsize { x: 0.25 y: 0.125 z: 0.5 }
      }
    }
    inertia { x: 1 y: 1 z: 1 }
    mass: 1.0
  }
  bodies {
    name: "Front Left Lower"
    colliders {
      box {
        halfsize { x: 0.25 y: 0.125 z: 0.5 }
      }
    }
    inertia { x: 1 y: 1 z: 1 }
    mass: 1.0
  }
  bodies {
    name: "Back Right Upper"
    colliders {
      box {
        halfsize { x: 0.25 y: 0.125 z: 0.5 }
      }
    }
    inertia { x: 1 y: 1 z: 1 }
    mass: 1.0
  }
  bodies {
    name: "Back Right Lower"
    colliders {
      box {
        halfsize { x: 0.25 y: 0.125 z: 0.5 }
      }
    }
    inertia { x: 1 y: 1 z: 1 }
    mass: 1.0
  }
  bodies {
    name: "Back Left Upper"
    colliders {
      box {
        halfsize { x: 0.25 y: 0.125 z: 0.5 }
      }
    }
    inertia { x: 1 y: 1 z: 1 }
    mass: 1.0
  }
  bodies {
    name: "Back Left Lower"
    colliders {
      box {
        halfsize { x: 0.25 y: 0.125 z: 0.5 }
      }
    }
    inertia { x: 1 y: 1 z: 1 }
    mass: 1.0
  }
  bodies {
    name: "Ground"
    colliders { plane {} }
    frozen { all: true }
  }
  bodies {
    name: "Target"
    colliders { sphere { radius: 2 }}
    frozen { all: true }
  }
  joints {
    name: "Torso_Shoulders"
    angle_limit { min: -60 max: 60 }
    parent_offset { x: 1.0 }
    child_offset {}
    parent: "Torso"
    child: "Shoulders"
    angular_damping: 35
  }
  joints {
    name: "Torso_Hips"
    angle_limit { min: -60 max: 60 }
    parent_offset { x: -1.0 }
    child_offset {}
    parent: "Torso"
    child: "Hips"
    angular_damping: 35
  }
  joints {
    name: "Shoulders_Front Right Upper"
    angle_limit { min: -60 max: 60 }
    rotation { z: 90 }
    parent_offset { y: -0.875 }
    child_offset { z: 0.375 }
    parent: "Shoulders"
    child: "Front Right Upper"
    angular_damping: 35
  }
  joints {
    name: "Front Right Upper_Front Right Lower"
    angle_limit { min: -60 max: 60 }
    rotation { z: 90 }
    parent_offset { y: 0.25 z: -0.25 }
    child_offset { z: 0.25 }
    parent: "Front Right Upper"
    child: "Front Right Lower"
    angular_damping: 35
  }
  joints {
    name: "Shoulders_Front Left Upper"
    angle_limit { min: -60 max: 60 }
    rotation { z: 90 }
    parent_offset { y: 0.875 }
    child_offset { z: 0.375 }
    parent: "Shoulders"
    child: "Front Left Upper"
    angular_damping: 35
  }
  joints {
    name: "Front Left Upper_Front Left Lower"
    angle_limit { min: -60 max: 60 }
    rotation { z: 90 }
    parent_offset { y: -0.25 z: -0.25 }
    child_offset { z: 0.25 }
    parent: "Front Left Upper"
    child: "Front Left Lower"
    angular_damping: 35
  }
  joints {
    name: "Hips_Back Right Upper"
    angle_limit { min: -60 max: 60 }
    rotation { z: 90 }
    parent_offset { y: -0.875 }
    child_offset { z: 0.375 }
    parent: "Hips"
    child: "Back Right Upper"
    angular_damping: 35
  }
  joints {
    name: "Back Right Upper_Back Right Lower"
    angle_limit { min: -60 max: 60 }
    rotation { z: 90 }
    parent_offset { y: 0.25 z: -0.25 }
    child_offset { z: 0.25 }
    parent: "Back Right Upper"
    child: "Back Right Lower"
    angular_damping: 35
  }
  joints {
    name: "Hips_Back Left Upper"
    angle_limit { min: -60 max: 60 }
    rotation { z: 90 }
    parent_offset { y: 0.875 }
    child_offset { z: 0.375 }
    parent: "Hips"
    child: "Back Left Upper"
    angular_damping: 35
  }
  joints {
    name: "Back Left Upper_Back Left Lower"
    angle_limit { min: -60 max: 60 }
    rotation { z: 90 }
    parent_offset { y: -0.25 z: -0.25 }
    child_offset { z: 0.25 }
    parent: "Back Left Upper"
    child: "Back Left Lower"
    angular_damping: 35
  }
  actuators {
    name: "Torso_Shoulders"
    torque {}
    joint: "Torso_Shoulders"
    strength: 300.0
  }
  actuators {
    name: "Torso_Hips"
    torque {}
    joint: "Torso_Hips"
    strength: 300.0
  }
  actuators {
    name: "Shoulders_Front Right Upper"
    torque {}
    joint: "Shoulders_Front Right Upper"
    strength: 300.0
  }
  actuators {
    name: "Front Right Upper_Front Right Lower"
    torque {}
    joint: "Front Right Upper_Front Right Lower"
    strength: 300.0
  }
  actuators {
    name: "Shoulders_Front Left Upper"
    torque {}
    joint: "Shoulders_Front Left Upper"
    strength: 300.0
  }
  actuators {
    name: "Front Left Upper_Front Left Lower"
    torque {}
    joint: "Front Left Upper_Front Left Lower"
    strength: 300.0
  }
  actuators {
    name: "Hips_Back Right Upper"
    torque {}
    joint: "Hips_Back Right Upper"
    strength: 300.0
  }
  actuators {
    name: "Back Right Upper_Back Right Lower"
    torque {}
    joint: "Back Right Upper_Back Right Lower"
    strength: 300.0
  }
  actuators {
    name: "Hips_Back Left Upper"
    torque {}
    joint: "Hips_Back Left Upper"
    strength: 300.0
  }
  actuators {
    name: "Back Left Upper_Back Left Lower"
    torque {}
    joint: "Back Left Upper_Back Left Lower"
    strength: 300.0
  }
  friction: 0.77459666924
  gravity { z: -9.8 }
  angular_damping: -0.05
  collide_include {
    first: "Front Right Lower"
    second: "Ground"
  }
  collide_include {
    first: "Front Left Lower"
    second: "Ground"
  }
  collide_include {
    first: "Back Right Lower"
    second: "Ground"
  }
  collide_include {
    first: "Back Left Lower"
    second: "Ground"
  }
  dt: 0.02
  substeps: 4
  """

_SYSTEM_CONFIG_SPRING = """
bodies {
  name: "Torso"
  colliders {
    box {
      halfsize { x: 0.75 y: 0.25 z: 0.125 }
    }
  }
  inertia { x: 1 y: 1 z: 1 }
  mass: 1.0
}
bodies {
  name: "Shoulders"
  colliders {
    box {
      halfsize { x: 0.25 y: 0.75 z: 0.125 }
    }
  }
  inertia { x: 1 y: 1 z: 1 }
  mass: 1.0
}
bodies {
  name: "Hips"
  colliders {
    box {
      halfsize { x: 0.25 y: 0.75 z: 0.125 }
    }
  }
  inertia { x: 1 y: 1 z: 1 }
  mass: 1.0
}
bodies {
  name: "Front Right Upper"
  colliders {
    box {
      halfsize { x: 0.25 y: 0.125 z: 0.5 }
    }
  }
  inertia { x: 1 y: 1 z: 1 }
  mass: 1.0
}
bodies {
  name: "Front Right Lower"
  colliders {
    box {
      halfsize { x: 0.25 y: 0.125 z: 0.5 }
    }
  }
  inertia { x: 1 y: 1 z: 1 }
  mass: 1.0
}

bodies {
  name: "Front Left Upper"
  colliders {
    box {
      halfsize { x: 0.25 y: 0.125 z: 0.5 }
    }
  }
  inertia { x: 1 y: 1 z: 1 }
  mass: 1.0
}
bodies {
  name: "Front Left Lower"
  colliders {
    box {
      halfsize { x: 0.25 y: 0.125 z: 0.5 }
    }
  }
  inertia { x: 1 y: 1 z: 1 }
  mass: 1.0
}
bodies {
  name: "Back Right Upper"
  colliders {
    box {
      halfsize { x: 0.25 y: 0.125 z: 0.5 }
    }
  }
  inertia { x: 1 y: 1 z: 1 }
  mass: 1.0
}
bodies {
  name: "Back Right Lower"
  colliders {
    box {
      halfsize { x: 0.25 y: 0.125 z: 0.5 }
    }
  }
  inertia { x: 1 y: 1 z: 1 }
  mass: 1.0
}
bodies {
  name: "Back Left Upper"
  colliders {
    box {
      halfsize { x: 0.25 y: 0.125 z: 0.5 }
    }
  }
  inertia { x: 1 y: 1 z: 1 }
  mass: 1.0
}
bodies {
  name: "Back Left Lower"
  colliders {
    box {
      halfsize { x: 0.25 y: 0.125 z: 0.5 }
    }
  }
  inertia { x: 1 y: 1 z: 1 }
  mass: 1.0
}
bodies {
  name: "Ground"
  colliders { plane {} }
  frozen { all: true }
}
bodies {
  name: "Target"
  colliders { sphere { radius: 2 }}
  frozen { all: true }
}
joints {
  name: "Torso_Shoulders"
  angle_limit { min: -60 max: 60 }
  parent_offset { x: 1.0 }
  child_offset {}
  parent: "Torso"
  child: "Shoulders"
  stiffness: 5000.0
  angular_damping: 35
}
joints {
  name: "Torso_Hips"
  angle_limit { min: -60 max: 60 }
  parent_offset { x: -1.0 }
  child_offset {}
  parent: "Torso"
  child: "Hips"
  stiffness: 5000.0
  angular_damping: 35
}
joints {
  name: "Shoulders_Front Right Upper"
  angle_limit { min: -60 max: 60 }
  rotation { z: 90 }
  parent_offset { y: -0.875 }
  child_offset { z: 0.375 }
  parent: "Shoulders"
  child: "Front Right Upper"
  stiffness: 5000.0
  angular_damping: 35
}
joints {
  name: "Front Right Upper_Front Right Lower"
  angle_limit { min: -60 max: 60 }
  rotation { z: 90 }
  parent_offset { y: 0.25 z: -0.25 }
  child_offset { z: 0.25 }
  parent: "Front Right Upper"
  child: "Front Right Lower"
  stiffness: 5000.0
  angular_damping: 35
}
joints {
  name: "Shoulders_Front Left Upper"
  angle_limit { min: -60 max: 60 }
  rotation { z: 90 }
  parent_offset { y: 0.875 }
  child_offset { z: 0.375 }
  parent: "Shoulders"
  child: "Front Left Upper"
  stiffness: 5000.0
  angular_damping: 35
}
joints {
  name: "Front Left Upper_Front Left Lower"
  angle_limit { min: -60 max: 60 }
  rotation { z: 90 }
  parent_offset { y: -0.25 z: -0.25 }
  child_offset { z: 0.25 }
  parent: "Front Left Upper"
  child: "Front Left Lower"
  stiffness: 5000.0
  angular_damping: 35
}
joints {
  name: "Hips_Back Right Upper"
  angle_limit { min: -60 max: 60 }
  rotation { z: 90 }
  parent_offset { y: -0.875 }
  child_offset { z: 0.375 }
  parent: "Hips"
  child: "Back Right Upper"
  stiffness: 5000.0
  angular_damping: 35
}
joints {
  name: "Back Right Upper_Back Right Lower"
  angle_limit { min: -60 max: 60 }
  rotation { z: 90 }
  parent_offset { y: 0.25 z: -0.25 }
  child_offset { z: 0.25 }
  parent: "Back Right Upper"
  child: "Back Right Lower"
  stiffness: 5000.0
  angular_damping: 35
}
joints {
  name: "Hips_Back Left Upper"
  angle_limit { min: -60 max: 60 }
  rotation { z: 90 }
  parent_offset { y: 0.875 }
  child_offset { z: 0.375 }
  parent: "Hips"
  child: "Back Left Upper"
  stiffness: 5000.0
  angular_damping: 35
}
joints {
  name: "Back Left Upper_Back Left Lower"
  angle_limit { min: -60 max: 60 }
  rotation { z: 90 }
  parent_offset { y: -0.25 z: -0.25 }
  child_offset { z: 0.25 }
  parent: "Back Left Upper"
  child: "Back Left Lower"
  stiffness: 5000.0
  angular_damping: 35
}
actuators {
  name: "Torso_Shoulders"
  torque {}
  joint: "Torso_Shoulders"
  strength: 300.0
}
actuators {
  name: "Torso_Hips"
  torque {}
  joint: "Torso_Hips"
  strength: 300.0
}
actuators {
  name: "Shoulders_Front Right Upper"
  torque {}
  joint: "Shoulders_Front Right Upper"
  strength: 300.0
}
actuators {
  name: "Front Right Upper_Front Right Lower"
  torque {}
  joint: "Front Right Upper_Front Right Lower"
  strength: 300.0
}
actuators {
  name: "Shoulders_Front Left Upper"
  torque {}
  joint: "Shoulders_Front Left Upper"
  strength: 300.0
}
actuators {
  name: "Front Left Upper_Front Left Lower"
  torque {}
  joint: "Front Left Upper_Front Left Lower"
  strength: 300.0
}
actuators {
  name: "Hips_Back Right Upper"
  torque {}
  joint: "Hips_Back Right Upper"
  strength: 300.0
}
actuators {
  name: "Back Right Upper_Back Right Lower"
  torque {}
  joint: "Back Right Upper_Back Right Lower"
  strength: 300.0
}
actuators {
  name: "Hips_Back Left Upper"
  torque {}
  joint: "Hips_Back Left Upper"
  strength: 300.0
}
actuators {
  name: "Back Left Upper_Back Left Lower"
  torque {}
  joint: "Back Left Upper_Back Left Lower"
  strength: 300.0
}
friction: 0.77459666924
gravity { z: -9.8 }
angular_damping: -0.05
baumgarte_erp: 0.1
collide_include {
  first: "Front Right Lower"
  second: "Ground"
}
collide_include {
  first: "Front Left Lower"
  second: "Ground"
}
collide_include {
  first: "Back Right Lower"
  second: "Ground"
}
collide_include {
  first: "Back Left Lower"
  second: "Ground"
}
dt: 0.02
substeps: 4
dynamics_mode: "legacy_spring"
"""



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Multi-agent RL utilities.

ComposerEnv() is inherently object-oriented (observations, actions,
rewards are all dictionaries), and therefore it is
straight-forward to enable multi-agent environments.

composer_env.metadata.agent_groups specifies full information for the
multi-agent task. E.g. envs/ma_descs.py.py.

E.g.
   agent_groups=dict(
                agent1=dict(reward_names=(('dist', 'agent1', 'agent2'),)),
                agent2=dict(reward_agents=('agent2',)),
            )
   means:
     - agent1 uses 'dist' reward function between 'agent1' and 'agent2'
     - agent2 uses all rewards of 'agent2'
     - by defaults, each agent uses its own action_space
       e.g. equivalent to agent1=(..., action_agents=('agent1',), ...)

agent_groups currently defines which rewards/actions belong to which agent.
observation is the same among all agents (TODO: add optionality).
"""

from collections import OrderedDict as odict
from typing import Dict, Tuple, Any, List
from brax.experimental.composer import component_editor as ce
from jax import numpy as jnp


def set_agent_groups(metadata: Any, action_shapes: Dict[str, Any],
                     observer_shapes: Dict[str, Any]):
  """Set metadata.agent_groups and return additional infos.

  Args:
    metadata: a Metadata object of ComposerEnv()
    action_shapes: an OrderedDict of sys action shape info (data_utils.py)
    observer_shapes: an OrderedDict of observation shape info (data_utils.py)

  Returns:
    group_action_shapes: an OrderedDict of agent-based action shape info
  """
  del observer_shapes
  if not metadata.agent_groups:
    return {}
  group_action_shapes = odict()
  for _, (k, v) in enumerate(sorted(metadata.agent_groups.items())):
    set_names_info(v, sorted(metadata.reward_fns.keys()), 'reward')
    set_names_info(v, action_shapes.keys(), 'action', default_agents=(k,))
    group_action_shapes[k] = get_action_shape(v, action_shapes)
  return group_action_shapes


def process_agent_rewards(metadata: Any,
                          reward_tuple_dict: Dict[str, Tuple[jnp.ndarray]]):
  """Process reward etc.

  based on metadata.agent_groups.

  Args:
    metadata: Metadata object in ComposerEnv()
    reward_tuple_dict: a dict of scalars (reward, score, done)

  Returns:
    reward: a jnp.ndarray of size [num_agents]
    score: a jnp.ndarray of size [num_agents]
    done: a jnp.ndarray of size [num_agents]
  """
  num_agents = len(metadata.agent_groups)
  reward, score, done = jnp.zeros((3,) + (num_agents,))
  all_reward_names = ()
  for i, (_, v) in enumerate(sorted(metadata.agent_groups.items())):
    reward_names = v.get('reward_names', ())
    for reward_name in reward_names:
      assert reward_name in reward_tuple_dict, (
          f'{reward_name} not in {reward_tuple_dict.keys()}')
      r, s, d = reward_tuple_dict[reward_name]
      reward = reward.at[i].add(r)
      score = score.at[i].add(s)
      done = done.at[i].set(jnp.logical_or(done[i], d))
    all_reward_names += reward_names
  assert set(all_reward_names) == set(reward_tuple_dict.keys()), (
      f'{set(all_reward_names)} != {set(reward_tuple_dict.keys())}')
  done = jnp.any(done, axis=-1)  # ensure done is a scalar
  return reward, score, done


def get_action_shape(v: Dict[str, Tuple[Any]], action_shapes: Dict[str, Any]):
  """Set action_indices."""
  names = v.get('action_names')
  indices = ()
  for name in names:
    s = action_shapes[name]
    indices_ = s.get('indices', list(range(s['start'], s['end'])))
    indices += tuple(indices_)
  return dict(size=len(indices), shape=(len(indices),), indices=indices)


def set_names_info(
    v: Dict[str, Tuple[Any]],
    all_names: List[str],
    var: str = 'reward',
    default_agents: Tuple[str] = (),
):
  """Set names based on '{var}_(names|agents)'."""
  names = v.get(f'{var}_names', ())
  assert all(isinstance(v, tuple)
             for v in names), f'{names} must be a Sequence of Tuples'
  names = tuple(ce.concat_name(*v) for v in names)
  agents = v.get(f'{var}_agents', default_agents)
  for agent in agents:
    agent_args = (agent,) if isinstance(agent, str) else agent
    assert isinstance(agent_args, (tuple, list)), agent_args
    names += tuple(k for k in all_names if ce.match_name(k, *agent_args))
  names = tuple(dict.fromkeys(names))  # remove duplicates/keep order
  v[f'{var}_names'] = names



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Brax training gradient utility functions."""

from typing import Callable, Optional

import jax
import optax


def loss_and_pgrad(loss_fn: Callable[..., float],
                   pmap_axis_name: Optional[str],
                   has_aux: bool = False):
  g = jax.value_and_grad(loss_fn, has_aux=has_aux)

  def h(*args, **kwargs):
    value, grad = g(*args, **kwargs)
    return value, jax.lax.pmean(grad, axis_name=pmap_axis_name)

  return g if pmap_axis_name is None else h


def gradient_update_fn(loss_fn: Callable[..., float],
                       optimizer: optax.GradientTransformation,
                       pmap_axis_name: Optional[str],
                       has_aux: bool = False):
  """Wrapper of the loss function that apply gradient updates.

  Args:
    loss_fn: The loss function.
    optimizer: The optimizer to apply gradients.
    pmap_axis_name: If relevant, the name of the pmap axis to synchronize
      gradients.
    has_aux: Whether the loss_fn has auxiliary data.

  Returns:
    A function that takes the same argument as the loss function plus the
    optimizer state. The output of this function is the loss, the new parameter,
    and the new optimizer state.
  """
  loss_and_pgrad_fn = loss_and_pgrad(
      loss_fn, pmap_axis_name=pmap_axis_name, has_aux=has_aux)

  def f(*args, optimizer_state):
    value, grads = loss_and_pgrad_fn(*args)
    params_update, optimizer_state = optimizer.update(grads, optimizer_state)
    params = optax.apply_updates(args[0], params_update)
    return value, params, optimizer_state

  return f



# Copyright 2022 The Brax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""RL training with an environment running entirely on an accelerator."""

import os

from absl import app
from absl import flags
from absl import logging
from brax import envs
from brax.io import html
from brax.io import metrics
from brax.io import model
from brax.io import npy_file
from brax.training.agents.apg import train as apg
from brax.training.agents.ars import train as ars
from brax.training.agents.es import train as es
from brax.training.agents.ppo import train as ppo
from brax.training.agents.sac import train as sac
import jax


FLAGS = flags.FLAGS

flags.DEFINE_enum('learner', 'ppo', ['ppo', 'apg', 'es', 'sac', 'ars'],
                  'Which algorithm to run.')
flags.DEFINE_string('env', 'ant', 'Name of environment to train.')
flags.DEFINE_integer('total_env_steps', 50000000,
                     'Number of env steps to run training for.')
flags.DEFINE_integer('num_evals', 10, 'How many times to run an eval.')
flags.DEFINE_integer('seed', 0, 'Random seed.')
flags.DEFINE_integer('num_envs', 4, 'Number of envs to run in parallel.')
flags.DEFINE_integer('action_repeat', 1, 'Action repeat.')
flags.DEFINE_integer('unroll_length', 30, 'Unroll length.')
flags.DEFINE_integer('batch_size', 4, 'Batch size.')
flags.DEFINE_integer('num_minibatches', 1, 'Number')
flags.DEFINE_integer(
    'num_updates_per_batch', 1,
    'Number of times to reuse each transition for gradient '
    'computation.')
flags.DEFINE_float('reward_scaling', 10.0, 'Reward scale.')
flags.DEFINE_float('entropy_cost', 3e-4, 'Entropy cost.')
flags.DEFINE_integer('episode_length', 1000, 'Episode length.')
flags.DEFINE_float('discounting', 0.99, 'Discounting.')
flags.DEFINE_float('learning_rate', 5e-4, 'Learning rate.')
flags.DEFINE_float('max_gradient_norm', 1e9,
                   'Maximal norm of a gradient update.')
flags.DEFINE_string('logdir', '', 'Logdir.')
flags.DEFINE_integer('num_rlds_episodes', 0,
                     'Number of episodes to output to RLDS.')

flags.DEFINE_bool('normalize_observations', True,
                  'Whether to apply observation normalization.')
flags.DEFINE_integer(
    'max_devices_per_host', None,
    'Maximum number of devices to use per host. If None, '
    'defaults to use as much as it can.')
flags.DEFINE_integer('num_videos', 1,
                     'Number of videos to record after training.')
flags.DEFINE_integer('num_trajectories_npy', 0,
                     'Number of rollouts to write to disk as raw QP states.')
# Evolution Strategy related flags
flags.DEFINE_integer(
    'population_size', 1,
    'Number of environments in ES. The actual number is 2x '
    'larger (used for antithetic sampling.')
flags.DEFINE_float('perturbation_std', 0.1,
                   'Std of a random noise added by ES.')
flags.DEFINE_enum('fitness_shaping', 'original',
                  ['original', 'centered_rank', 'wierstra'],
                  'Defines a type of fitness shaping to apply.')
flags.DEFINE_bool('center_fitness', False,
                  'Whether to normalize fitness after the shaping.')
flags.DEFINE_float('l2coeff', 0,
                   'L2 regularization coefficient for model params.')
# SAC hps.
flags.DEFINE_integer('min_replay_size', 8192,
                     'Minimal replay buffer size before the training starts.')
flags.DEFINE_integer('max_replay_size', 1048576, 'Maximal replay buffer size.')
flags.DEFINE_integer(
    'grad_updates_per_step', 1,
    'How many SAC gradient updates to run per one step in the '
    'environment.')
# PPO hps.
flags.DEFINE_float('gae_lambda', .95, 'General advantage estimation lambda.')
flags.DEFINE_float('clipping_epsilon', .3, 'Policy loss clipping epsilon.')
# ARS hps.
flags.DEFINE_integer(
    'number_of_directions', 60,
    'Number of directions to explore. The actual number is 2x '
    'larger (used for antithetic sampling.')
flags.DEFINE_integer('top_directions', 20,
                     'Number of top directions to select.')
flags.DEFINE_float('exploration_noise_std', 0.1,
                   'Std of a random noise added by ARS.')
flags.DEFINE_float('reward_shift', 0.,
                   'A reward shift to get rid of "stay alive" bonus.')
flags.DEFINE_enum('head_type', '', ['', 'clip', 'tanh'],
                  'Which policy head to use.')
# ARS hps.
flags.DEFINE_integer('truncation_length', None,
                     'Truncation for gradient propagation in APG.')


def main(unused_argv):

  with metrics.Writer(FLAGS.logdir) as writer:
    writer.write_hparams({
        'num_evals': FLAGS.num_evals,
        'num_envs': FLAGS.num_envs,
        'total_env_steps': FLAGS.total_env_steps
    })
    if FLAGS.learner == 'sac':
      make_policy, params, _ = sac.train(
          environment=envs.get_environment(FLAGS.env),
          num_envs=FLAGS.num_envs,
          action_repeat=FLAGS.action_repeat,
          normalize_observations=FLAGS.normalize_observations,
          num_timesteps=FLAGS.total_env_steps,
          num_evals=FLAGS.num_evals,
          batch_size=FLAGS.batch_size,
          min_replay_size=FLAGS.min_replay_size,
          max_replay_size=FLAGS.max_replay_size,
          learning_rate=FLAGS.learning_rate,
          discounting=FLAGS.discounting,
          max_devices_per_host=FLAGS.max_devices_per_host,
          seed=FLAGS.seed,
          reward_scaling=FLAGS.reward_scaling,
          grad_updates_per_step=FLAGS.grad_updates_per_step,
          episode_length=FLAGS.episode_length,
          progress_fn=writer.write_scalars)
    if FLAGS.learner == 'es':
      make_policy, params, _ = es.train(
          environment=envs.get_environment(FLAGS.env),
          num_timesteps=FLAGS.total_env_steps,
          fitness_shaping=es.FitnessShaping[FLAGS.fitness_shaping.upper()],
          population_size=FLAGS.population_size,
          perturbation_std=FLAGS.perturbation_std,
          normalize_observations=FLAGS.normalize_observations,
          action_repeat=FLAGS.action_repeat,
          num_evals=FLAGS.num_evals,
          center_fitness=FLAGS.center_fitness,
          l2coeff=FLAGS.l2coeff,
          learning_rate=FLAGS.learning_rate,
          seed=FLAGS.seed,
          max_devices_per_host=FLAGS.max_devices_per_host,
          episode_length=FLAGS.episode_length,
          progress_fn=writer.write_scalars)
    if FLAGS.learner == 'ppo':
      make_policy, params, _ = ppo.train(
          environment=envs.get_environment(FLAGS.env),
          num_timesteps=FLAGS.total_env_steps,
          episode_length=FLAGS.episode_length,
          action_repeat=FLAGS.action_repeat,
          num_envs=FLAGS.num_envs,
          max_devices_per_host=FLAGS.max_devices_per_host,
          learning_rate=FLAGS.learning_rate,
          entropy_cost=FLAGS.entropy_cost,
          discounting=FLAGS.discounting,
          seed=FLAGS.seed,
          unroll_length=FLAGS.unroll_length,
          batch_size=FLAGS.batch_size,
          num_minibatches=FLAGS.num_minibatches,
          normalize_observations=FLAGS.normalize_observations,
          num_updates_per_batch=FLAGS.num_updates_per_batch,
          num_evals=FLAGS.num_evals,
          reward_scaling=FLAGS.reward_scaling,
          gae_lambda=FLAGS.gae_lambda,
          clipping_epsilon=FLAGS.clipping_epsilon,
          progress_fn=writer.write_scalars)
    if FLAGS.learner == 'apg':
      make_policy, params, _ = apg.train(
          environment=envs.get_environment(FLAGS.env),
          num_envs=FLAGS.num_envs,
          action_repeat=FLAGS.action_repeat,
          num_evals=FLAGS.num_evals,
          learning_rate=FLAGS.learning_rate,
          seed=FLAGS.seed,
          max_devices_per_host=FLAGS.max_devices_per_host,
          normalize_observations=FLAGS.normalize_observations,
          max_gradient_norm=FLAGS.max_gradient_norm,
          episode_length=FLAGS.episode_length,
          truncation_length=FLAGS.truncation_length,
          progress_fn=writer.write_scalars)
    if FLAGS.learner == 'ars':
      make_policy, params, _ = ars.train(
          environment=envs.get_environment(FLAGS.env),
          number_of_directions=FLAGS.number_of_directions,
          max_devices_per_host=FLAGS.max_devices_per_host,
          action_repeat=FLAGS.action_repeat,
          normalize_observations=FLAGS.normalize_observations,
          num_timesteps=FLAGS.total_env_steps,
          exploration_noise_std=FLAGS.exploration_noise_std,
          num_evals=FLAGS.num_evals,
          seed=FLAGS.seed,
          step_size=FLAGS.learning_rate,
          top_directions=FLAGS.top_directions,
          reward_shift=FLAGS.reward_shift,
          episode_length=FLAGS.episode_length,
          progress_fn=writer.write_scalars)

  # Save to flax serialized checkpoint.
  filename = f'{FLAGS.env}_{FLAGS.learner}.pkl'
  path = os.path.join(FLAGS.logdir, filename)
  model.save_params(path, params)

  # Output an episode trajectory.
  env = envs.create(FLAGS.env)

  @jax.jit
  def jit_next_state(state, key):
    new_key, tmp_key = jax.random.split(key)
    act = make_policy(params)(state.obs, tmp_key)[0]
    return env.step(state, act), act, new_key

  def do_rollout(rng):
    rng, env_key = jax.random.split(rng)
    state = env.reset(env_key)
    qps = []
    while not state.done:
      qps.append(state.qp)
      state, _, rng = jit_next_state(state, rng)
    return qps, rng

  trajectories = []
  rng = jax.random.PRNGKey(FLAGS.seed)
  for _ in range(max(FLAGS.num_videos, FLAGS.num_trajectories_npy)):
    qps, rng = do_rollout(rng)
    trajectories.append(qps)

  if hasattr(env, 'sys'):
    for i in range(FLAGS.num_videos):
      html_path = f'{FLAGS.logdir}/saved_videos/trajectory_{i:04d}.html'
      html.save_html(html_path, env.sys, trajectories[i], make_dir=True)
  elif FLAGS.num_videos > 0:
    logging.warn('Cannot save videos for non physics environments.')

  for i in range(FLAGS.num_trajectories_npy):
    qp_path = f'{FLAGS.logdir}/saved_qps/trajectory_{i:04d}.npy'
    npy_file.save(qp_path, trajectories[i], make_dir=True)



if __name__ == '__main__':
  app.run(main)
